#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import VariableNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleOp
from xrlint.util.schema import schema

DEFAULT_ATTRS = ["standard_name", "long_name"]


@plugin.define_rule(
    "var-desc",
    version="1.0.0",
    type="suggestion",
    description=(
        "Check that each data variable provides an"
        " identification and description of the content."
        " The rule can be configured by parameter `attrs` which is a list"
        " of names of attributes that provides descriptive information."
        f" It defaults to `{DEFAULT_ATTRS}`."
        ""
    ),
    docs_url=(
        "https://cfconventions.org/cf-conventions/cf-conventions.html#standard-name"
    ),
    schema=schema(
        "object",
        properties={
            "attrs": schema(
                "array",
                items=schema("string"),
                default=DEFAULT_ATTRS,
                title="Attribute names to check",
            ),
        },
    ),
)
class VarDesc(RuleOp):
    def __init__(self, attrs: list[str] | None = None):
        self._attrs = attrs if attrs is not None else DEFAULT_ATTRS

    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        if node.name not in ctx.dataset.data_vars:
            # This rule applies to data variables only
            return

        var_attrs = node.array.attrs
        for attr_name in self._attrs:
            if attr_name not in var_attrs:
                ctx.report(f"Missing attribute {attr_name!r}.")



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.config import ConfigObject
from xrlint.formatter import FormatterContext
from xrlint.plugin import new_plugin
from xrlint.result import Message, Result, ResultStats
from xrlint.rule import RuleOp


class FormatterContextImpl(FormatterContext):
    def __init__(self, max_warnings: int = -1):
        self._max_warnings = max_warnings
        self._result_stats = ResultStats()

    @property
    def max_warnings_exceeded(self) -> bool:
        return self._result_stats.warning_count > self._max_warnings

    @property
    def result_stats(self) -> ResultStats:
        return self._result_stats


def get_context(max_warnings: int = -1) -> FormatterContext:
    return FormatterContextImpl(max_warnings)


def get_test_results():
    plugin = new_plugin(name="test")

    @plugin.define_rule(
        "rule-1", description="Haha", docs_url="https://rules.com/haha.html"
    )
    class Rule1(RuleOp):
        pass

    @plugin.define_rule(
        "rule-2", description="Hoho", docs_url="https://rules.com/hoho.html"
    )
    class Rule2(RuleOp):
        pass

    config_obj = ConfigObject(plugins={"test": plugin})

    return [
        Result(
            file_path="test.nc",
            config_object=config_obj,
            messages=[
                Message(
                    message="message-1",
                    rule_id="test/rule-1",
                    severity=2,
                    node_path="dataset",
                ),
                Message(
                    message="message-2",
                    rule_id="test/rule-2",
                    severity=1,
                    node_path="dataset",
                ),
                Message(message="message-3", fatal=True),
            ],
        ),
        Result(
            file_path="test.nc",
            config_object=config_obj,
            messages=[
                Message(message="message-1", rule_id="test/rule-1", severity=1),
                Message(message="message-2", rule_id="test/rule-2", severity=2),
                Message(message="message-3", fatal=False),
            ],
        ),
    ]



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import VariableNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "var-units",
    version="1.0.0",
    type="suggestion",
    description="Every variable should provide a description of its units.",
    docs_url="https://cfconventions.org/cf-conventions/cf-conventions.html#units",
)
class VarUnits(RuleOp):
    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        array = node.array
        attrs = array.attrs

        if "grid_mapping_name" in attrs:
            # likely grid mapping variable --> rule "gid-mappings"
            return
        if "units" in array.encoding:
            # likely time coordinate --> rule "time-coordinate"
            return

        units = attrs.get("units")
        if "units" not in attrs:
            ctx.report("Missing attribute 'units'.")
        elif not isinstance(units, str):
            ctx.report(f"Invalid attribute 'units': {units!r}")
        elif not units:
            ctx.report("Empty attribute 'units'.")



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from abc import ABC
from dataclasses import dataclass
from typing import Type
from unittest import TestCase

import pytest

from xrlint.operation import Operation, OperationMeta


class ThingOp(ABC):
    def do_something(self):
        pass


@dataclass(kw_only=True)
class ThingMeta(OperationMeta):
    pass


@dataclass(kw_only=True, frozen=True)
class Thing(Operation):
    meta: ThingMeta
    op_class: Type[ThingOp]

    @classmethod
    def meta_class(cls) -> Type:
        return ThingMeta

    @classmethod
    def op_base_class(cls) -> Type[ThingOp]:
        return ThingOp

    @classmethod
    def value_name(cls) -> str:
        return "thing"

    @classmethod
    def define(cls, op_class: Type[ThingOp] | None = None, **kwargs):
        return cls.define_operation(op_class, **kwargs)


class MyThingOp1(ThingOp):
    pass


class MyThingOp2(ThingOp):
    pass


# This is one way of exporting one of many things from a module
thing1 = Thing(meta=ThingMeta(name="my-thing-op-1"), op_class=MyThingOp1)
thing2 = Thing(meta=ThingMeta(name="my-thing-op-2"), op_class=MyThingOp2)


# This is the default way of exporting a single thing from a dedicated module
def export_thing() -> Thing:
    class MyThingOp3(ThingOp):
        pass

    return Thing(meta=ThingMeta(name="my-thing-op-3"), op_class=MyThingOp3)


class OperationTest(TestCase):
    def test_class_props(self):
        self.assertEqual(OperationMeta, Operation.meta_class())
        self.assertEqual(type, Operation.op_base_class())
        self.assertEqual("export_operation", Operation.op_import_attr_name())
        self.assertEqual("operation", Operation.value_name())
        self.assertEqual(
            "Operation | Type[type] | dict | str", Operation.value_type_name()
        )

    def test_from_value_ok_rule(self):
        thing1_ = Thing.from_value(thing1)
        self.assertIs(thing1_, thing1)

    def test_from_value_ok_rule_op(self):
        class MyThingOp3(ThingOp):
            pass

        meta = ThingMeta(name="my-thing-op-3")

        # This "defines" MyThingOp2 so we can create
        # instances from its operation class.
        MyThingOp3.meta = meta

        thing3 = Thing.from_value(MyThingOp3)
        self.assertIsInstance(thing3, Thing)
        self.assertIs(meta, thing3.meta)
        self.assertIs(MyThingOp3, thing3.op_class)

    def test_from_value_ok_str(self):
        thing1_ = Thing.from_value("tests.test_operation:thing1")
        self.assertIs(thing1, thing1_)
        self.assertEqual("tests.test_operation:thing1", thing1_.meta.ref)

        thing2_ = Thing.from_value("tests.test_operation:thing2")
        self.assertIs(thing2, thing2_)
        self.assertEqual("tests.test_operation:thing2", thing2_.meta.ref)

        # default attribute is "export_thing"
        thing3 = Thing.from_value("tests.test_operation")
        self.assertIsInstance(thing3, Thing)
        self.assertIs("my-thing-op-3", thing3.meta.name)
        self.assertIsInstance(thing3.op_class, type)
        self.assertEqual("tests.test_operation:export_thing", thing3.meta.ref)

    # noinspection PyMethodMayBeStatic
    def test_from_value_fails(self):
        with pytest.raises(
            TypeError, match="value must be of type Thing | str, but got int"
        ):
            Thing.from_value(73)

        class MyThing3(ThingOp):
            """This is my 3rd thing."""

        with pytest.raises(
            ValueError,
            match=r"missing thing metadata, apply define_thing\(\) to class MyThing3",
        ):
            Thing.from_value(MyThing3)

        with pytest.raises(
            TypeError,
            match=(
                r"thing must be of type Thing \| Type\[ThingOp\] \|"
                r" dict | str, but got type"
            ),
        ):
            Thing.from_value(Thing)

    def test_to_json(self):
        class MyThingOp3(ThingOp):
            """This is my 3rd thing."""

        thing3 = Thing(
            meta=ThingMeta(name="t3", ref="mypkg.things:thing3"), op_class=MyThingOp3
        )
        self.assertEqual("mypkg.things:thing3", thing3.to_json())

        rule = Thing(
            meta=ThingMeta(name="t3", description="What a thing."), op_class=MyThingOp3
        )
        self.assertEqual(
            {
                "meta": {"description": "What a thing.", "name": "t3"},
                "op_class": "<class "
                "'tests.test_operation.OperationTest.test_to_json.<locals>.MyThingOp3'>",
            },
            rule.to_json(),
        )


class OpMixinDefineTest(TestCase):
    def test_define_op(self):
        class MyThingOp3(ThingOp):
            """This is my 3rd thing."""

        value = Thing.define_operation(MyThingOp3, meta_kwargs=dict(version="1.0"))
        self.assertIsInstance(value, Thing)
        self.assertIsInstance(value.meta, ThingMeta)
        self.assertEqual("my-thing-op-3", value.meta.name)
        self.assertEqual("1.0", value.meta.version)
        self.assertEqual("This is my 3rd thing.", value.meta.description)
        self.assertIs(MyThingOp3, value.op_class)
        self.assertTrue(hasattr(MyThingOp3, "meta"))
        # noinspection PyUnresolvedReferences
        self.assertIs(value.meta, MyThingOp3.meta)

    def test_define_op_fail(self):
        class MyThingOp3(ThingOp):
            """This is my 3rd thing."""

        with pytest.raises(
            TypeError, match="registry must be a MutableMapping, but got int"
        ):
            # noinspection PyTypeChecker
            Thing.define_operation(MyThingOp3, registry=12)

    def test_decorator(self):
        class MyThingOp3(ThingOp):
            """This is my 3rd thing."""

        closure = Thing.define()
        self.assertTrue(callable(closure))
        op_class = closure(MyThingOp3)
        self.assertIs(MyThingOp3, op_class)
        self.assertTrue(hasattr(MyThingOp3, "meta"))
        # noinspection PyUnresolvedReferences
        meta = op_class.meta
        self.assertEqual("my-thing-op-3", meta.name)
        self.assertEqual("0.0.0", meta.version)
        self.assertEqual("This is my 3rd thing.", meta.description)

    # noinspection PyMethodMayBeStatic
    def test_decorator_fail(self):
        closure = Thing.define()
        with pytest.raises(
            TypeError, match="decorated thing component must be a class, but got int"
        ):
            closure(32)

        with pytest.raises(
            TypeError,
            match=(
                "decorated thing component must be a subclass of ThingOp, but got Thing"
            ),
        ):
            closure(Thing)

    def test_function(self):
        class MyThingOp3(ThingOp):
            """This is my 3rd thing."""

        thing = Thing.define(op_class=MyThingOp3)
        self.assertIsInstance(thing, Thing)
        self.assertIs(MyThingOp3, thing.op_class)
        self.assertTrue(hasattr(MyThingOp3, "meta"))
        meta = thing.meta
        # noinspection PyUnresolvedReferences
        self.assertIs(meta, thing.op_class.meta)
        self.assertEqual("my-thing-op-3", meta.name)
        self.assertEqual("0.0.0", meta.version)
        self.assertEqual("This is my 3rd thing.", meta.description)

    # noinspection PyMethodMayBeStatic
    def test_function_fail(self):
        class MyThingOp3(ThingOp):
            """This is my 3rd thing."""

        with pytest.raises(TypeError, match="op_class must be a class, but got str"):
            # noinspection PyTypeChecker
            Thing.define(op_class="Huh!")

        with pytest.raises(
            TypeError,
            match="op_class must be a subclass of ThingOp, but got TestCase",
        ):
            # noinspection PyTypeChecker
            Thing.define(TestCase)

    def test_with_registry(self):
        class Op1(ThingOp):
            """This is my 3rd thing."""

        class Op2(ThingOp):
            """This is my 3rd thing."""

        registry = {}
        t1 = Thing.define(op_class=Op1, registry=registry)
        t2 = Thing.define(op_class=Op2, registry=registry)
        self.assertIs(t1, registry["op-1"])
        self.assertIs(t2, registry["op-2"])



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Any
from unittest import TestCase

import xarray as xr

from xrlint.plugin import Plugin, PluginMeta, new_plugin
from xrlint.processor import Processor, ProcessorOp
from xrlint.result import Message
from xrlint.rule import Rule, RuleOp, define_rule


class PluginTest(TestCase):
    def test_class_props(self):
        self.assertEqual("plugin", Plugin.value_name())
        self.assertEqual("Plugin | dict | str", Plugin.value_type_name())

    def test_new_plugin(self):
        plugin = new_plugin(name="hello", version="2.4.5")
        self.assertEqual(Plugin(meta=PluginMeta(name="hello", version="2.4.5")), plugin)

    def test_from_value_ok_plugin(self):
        plugin = Plugin(meta=PluginMeta(name="hello"))
        self.assertIs(plugin, Plugin.from_value(plugin))

    def test_from_value_ok_dict(self):
        @define_rule()
        class MyRule1(RuleOp):
            """This is my 1st rule."""

        @define_rule()
        class MyRule2(RuleOp):
            """This is my 2nd rule."""

        plugin = Plugin.from_value(
            {
                "meta": {
                    "name": "hello",
                    "version": "1.2.3",
                },
                "rules": {
                    "r1": MyRule1,
                    "r2": MyRule2,
                },
                "configs": {
                    "recommended": [
                        {
                            "rules": {
                                "hello/r1": "warn",
                                "hello/r2": "error",
                            },
                        }
                    ],
                },
            }
        )
        self.assertIsInstance(plugin, Plugin)


class PluginMetaTest(TestCase):
    def test_class_props(self):
        self.assertEqual("plugin_meta", PluginMeta.value_name())
        self.assertEqual("PluginMeta | dict", PluginMeta.value_type_name())

    def test_from_value(self):
        self.assertEqual(
            PluginMeta(name="p", ref="a.b.c:p"),
            PluginMeta.from_value({"name": "p", "version": "0.0.0", "ref": "a.b.c:p"}),
        )


class PluginDefineRuleDecoratorTest(TestCase):
    # noinspection PyUnusedLocal
    def test_decorator(self):
        plugin = Plugin(meta=PluginMeta(name="test"))

        @plugin.define_rule("my-rule-1")
        class MyRule1(RuleOp):
            pass

        @plugin.define_rule("my-rule-2")
        class MyRule2(RuleOp):
            pass

        @plugin.define_rule("my-rule-3")
        class MyRule3(RuleOp):
            pass

        rules = plugin.rules
        rule_names = list(rules.keys())
        rule1, rule2, rule3 = list(rules.values())
        self.assertEqual(["my-rule-1", "my-rule-2", "my-rule-3"], rule_names)
        self.assertIsInstance(rule1, Rule)
        self.assertIsInstance(rule2, Rule)
        self.assertIsInstance(rule3, Rule)
        self.assertIsNot(rule2, rule1)
        self.assertIsNot(rule3, rule1)
        self.assertIsNot(rule3, rule2)

        my_rule = plugin.rules.get("my-rule-1")
        self.assertIsInstance(my_rule, Rule)
        self.assertEqual("my-rule-1", my_rule.meta.name)
        self.assertEqual("0.0.0", my_rule.meta.version)
        self.assertEqual(None, my_rule.meta.schema)
        self.assertEqual("problem", my_rule.meta.type)


class PluginDefineProcessorDecoratorTest(TestCase):
    # noinspection PyUnusedLocal
    def test_decorator(self):
        plugin = Plugin(meta=PluginMeta(name="test"))

        @plugin.define_processor("my-processor-1")
        class MyProcessor1(ProcessorOp):
            def preprocess(
                self, file_path: str, opener_options: dict[str, Any]
            ) -> list[tuple[xr.Dataset, str]]:
                return []

            def postprocess(
                self, messages: list[list[Message]], file_path: str
            ) -> list[Message]:
                return []

        @plugin.define_processor("my-processor-2")
        class MyProcessor2(ProcessorOp):
            def preprocess(
                self, file_path: str, opener_options: dict[str, Any]
            ) -> list[tuple[xr.Dataset, str]]:
                return []

            def postprocess(
                self, messages: list[list[Message]], file_path: str
            ) -> list[Message]:
                return []

        processors = plugin.processors
        processors_names = list(processors.keys())
        processor1, processor2 = list(processors.values())
        self.assertEqual(["my-processor-1", "my-processor-2"], processors_names)
        self.assertIsInstance(processor1, Processor)
        self.assertIsInstance(processor2, Processor)
        self.assertIsNot(processor1, processor2)

        my_processor = plugin.processors.get("my-processor-1")
        self.assertIsInstance(my_processor, Processor)
        self.assertEqual("my-processor-1", my_processor.meta.name)
        self.assertEqual("0.0.0", my_processor.meta.version)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from os import PathLike
from pathlib import Path
from typing import Any

import xarray as xr

from xrlint.config import Config, ConfigLike, get_core_config_object
from xrlint.result import Result

from ._linter.validate import new_fatal_message, validate_dataset
from .constants import MISSING_DATASET_FILE_PATH, MISSING_DATATREE_FILE_PATH


def new_linter(*configs: ConfigLike, **config_props: Any) -> "Linter":
    """Create a new `Linter` with the core plugin included and the
     given additional configuration.

    Args:
        *configs: Variable number of configuration-like arguments.
            For more information see the
            [ConfigLike][xrlint.config.ConfigLike] type alias.
        **config_props: Individual configuration object properties.
            For more information refer to the properties of a
            [ConfigObject][xrlint.config.ConfigObject].

    Returns:
        A new linter instance
    """
    return Linter(get_core_config_object(), *configs, **config_props)


class Linter:
    """The linter.

    Using the constructor directly creates an empty linter
    with no configuration - even without the core plugin and
    its predefined rule configurations.
    If you want a linter with core plugin included use the
    `new_linter()` function.

    Args:
        *configs: Variable number of configuration-like arguments.
            For more information see the
            [ConfigLike][xrlint.config.ConfigLike] type alias.
        **config_props: Individual configuration object properties.
            For more information refer to the properties of a
            [ConfigObject][xrlint.config.ConfigObject].
    """

    def __init__(self, *configs: ConfigLike, **config_props: Any):
        self._config = Config.from_config(*configs, config_props)

    @property
    def config(self) -> Config:
        """Get this linter's configuration."""
        return self._config

    def validate(
        self,
        dataset: Any,
        *,
        file_path: str | None = None,
        config: ConfigLike = None,
        **config_props: Any,
    ) -> Result:
        """Validate a dataset against applicable rules.

        Args:
            dataset: The dataset. Can be a `xr.Dataset` or `xr.DataTree`
                instance or a file path, or any dataset source that can
                be opened using `xarray.open_dataset()`
                or `xarray.open_datatree()`.
            file_path: Optional file path used for formatting
                messages. Useful if `dataset` is not a file path.
            config: Optional configuration-like value.
                For more information see the
                [ConfigLike][xrlint.config.ConfigLike] type alias.
            **config_props: Individual configuration object properties.
                For more information refer to the properties of a
                [ConfigObject][xrlint.config.ConfigObject].

        Returns:
            Result of the validation.
        """
        if not file_path:
            if isinstance(dataset, (xr.Dataset, xr.DataTree)):
                file_path = file_path or _get_file_path_for_dataset(dataset)
            else:
                file_path = file_path or _get_file_path_for_source(dataset)

        config = Config.from_config(self._config, config, config_props)
        config_obj = config.compute_config_object(file_path)
        if config_obj is None or not config_obj.rules:
            return Result(
                file_path=file_path,
                messages=[
                    new_fatal_message(
                        f"No configuration given or matches {file_path!r}.",
                    )
                ],
            )

        return validate_dataset(config_obj, dataset, file_path)


def _get_file_path_for_dataset(dataset: xr.Dataset | xr.DataTree) -> str:
    ds_source = dataset.encoding.get("source")
    return _get_file_path_for_source(
        ds_source,
        MISSING_DATASET_FILE_PATH
        if isinstance(dataset, xr.Dataset)
        else MISSING_DATATREE_FILE_PATH,
    )


def _get_file_path_for_source(ds_source: Any, default: str | None = None) -> str:
    file_path = str(ds_source) if isinstance(ds_source, (str, Path, PathLike)) else ""
    return file_path or (default or MISSING_DATASET_FILE_PATH)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.plugins.core import export_plugin


class ExportPluginTest(TestCase):
    def test_rules_complete(self):
        plugin = export_plugin()
        self.assertEqual(
            {
                "access-latency",
                "content-desc",
                "conventions",
                "coords-for-dims",
                "grid-mappings",
                "lat-coordinate",
                "lon-coordinate",
                "no-empty-attrs",
                "no-empty-chunks",
                "time-coordinate",
                "var-desc",
                "var-flags",
                "var-missing-data",
                "var-units",
            },
            set(plugin.rules.keys()),
        )

    def test_configs_complete(self):
        plugin = export_plugin()
        self.assertEqual(
            {
                "all",
                "recommended",
            },
            set(plugin.configs.keys()),
        )
        all_rule_names = set(plugin.rules.keys())
        self.assertEqual(
            all_rule_names,
            set(plugin.configs["all"][-1].rules.keys()),
        )
        self.assertEqual(
            all_rule_names,
            set(plugin.configs["recommended"][-1].rules.keys()),
        )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

"""
This configuration example shows how to define and use a plugin
using the `Plugin` class and its `define_rule()` decorator method.

You can use this example directly via the Python API by passing it's
exported configuration to an instance of the `Linter` class or use
the XRLint CLI:

```bash
xrlint -c examples/plugin_config.py <OPTIONS> <FILES>
```
"""

from xrlint.node import DatasetNode
from xrlint.plugin import new_plugin
from xrlint.rule import RuleContext, RuleOp

plugin = new_plugin(name="hello-plugin", version="1.0.0")


@plugin.define_rule("good-title")
class GoodTitle(RuleOp):
    """Dataset title should be 'Hello World!'."""

    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        good_title = "Hello World!"
        if node.dataset.attrs.get("title") != good_title:
            ctx.report(
                "Attribute 'title' wrong.",
                suggestions=[f"Rename it to {good_title!r}."],
            )


# Define more rules here...


plugin.define_config(
    "recommended",
    [
        {
            "rules": {
                "hello/good-title": "warn",
                # Configure more rules here...
            },
        }
    ],
)

# Add more configurations here...


def export_config():
    return [
        # Use "hello" plugin
        {
            "plugins": {
                "hello": plugin,
            },
        },
        # Use recommended settings from xrlint
        "recommended",
        # Use recommended settings from "hello" plugin
        "hello/recommended",
    ]



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

"""
This example demonstrates how to develop new rules.
"""

import xarray as xr

from xrlint.node import DatasetNode
from xrlint.rule import RuleContext, RuleOp, define_rule
from xrlint.testing import RuleTest, RuleTester


# ----------------------------------------------------
# Place the rule implementation code in its own module
# ----------------------------------------------------


@define_rule("good-title")
class GoodTitle(RuleOp):
    """Dataset title should be 'Hello World!'."""

    # We just validate the dataset instance here. You can also implement
    # the validation of other nodes in this class, e.g.,
    # validate_datatree(), validate_variable(), validate_attrs(),
    # and validate_attr().
    #
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        good_title = "Hello World!"
        if node.dataset.attrs.get("title") != good_title:
            ctx.report(
                "Attribute 'title' wrong.",
                suggestions=[f"Rename it to {good_title!r}."],
            )


# ---------------------------------------------------
# Place the following rule test code in a test module
# ---------------------------------------------------

tester = RuleTester()

valid_dataset = xr.Dataset(attrs=dict(title="Hello World!"))
invalid_dataset = xr.Dataset(attrs=dict(title="Hello Hamburg!"))

# You can use the tester to run a test directly
#
tester.run(
    "good-title",
    GoodTitle,
    valid=[RuleTest(dataset=valid_dataset)],
    # We expect one message to be emitted
    invalid=[RuleTest(dataset=invalid_dataset, expected=1)],
)

# ... or generate a test class that will be derived from `unitest.TestCase`.
# This will provide you tooling support via your test runner, e.g., pytest,
# as the tests in `valid` and `invalid` will be transformed into
# test methods of the generated class.
#
GoodTitleTest = tester.define_test(
    "good-title",
    GoodTitle,
    valid=[RuleTest(dataset=valid_dataset)],
    # Note, here we expect a specific message to be emitted
    invalid=[RuleTest(dataset=invalid_dataset, expected=["Attribute 'title' wrong."])],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import DatasetNode
from xrlint.plugins.xcube.constants import GM_NAMES, GM_NAMES_TEXT
from xrlint.plugins.xcube.plugin import plugin
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "grid-mapping-naming",
    version="1.0.0",
    type="suggestion",
    description=(
        f"Grid mapping variables should be called {GM_NAMES_TEXT}"
        f" for compatibility with rioxarray and other packages."
    ),
    docs_url="https://xcube.readthedocs.io/en/latest/cubespec.html#spatial-reference",
)
class GridMappingNaming(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        for var_name, var in node.dataset.variables.items():
            if "grid_mapping_name" in var.attrs and var_name not in GM_NAMES:
                ctx.report(
                    f"Grid mapping variables should be named"
                    f" {GM_NAMES_TEXT},"
                    f" but name is {var_name!r}."
                )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import xarray as xr

from xrlint.plugins.core.rules.content_desc import ContentDesc
from xrlint.testing import RuleTest, RuleTester

global_attrs = dict(
    title="OC-Climatology",
    history="2025-01-26: created",
)

common_attrs = dict(
    institution="ESA",
    source="a.nc; b.nc",
    references="!",
    comment="?",
)

all_attrs = global_attrs | common_attrs

time_coord = xr.DataArray(
    [1, 2, 3], dims="time", attrs=dict(units="days since 2025-01-01")
)

valid_dataset_0 = xr.Dataset(
    attrs=all_attrs,
    data_vars=dict(chl=xr.DataArray([1, 2, 3], dims="time", attrs=dict())),
    coords=dict(time=time_coord),
)
valid_dataset_1 = xr.Dataset(
    attrs=global_attrs,
    data_vars=dict(chl=xr.DataArray([1, 2, 3], dims="time", attrs=common_attrs)),
    coords=dict(time=time_coord),
)
valid_dataset_1a = xr.Dataset(
    attrs=global_attrs,
    data_vars=dict(
        chl=xr.DataArray([1, 2, 3], dims="time", attrs=common_attrs),
        crs=xr.DataArray(0, attrs=dict(grid_mapping_name="...")),
    ),
    coords=dict(time=time_coord),
)
valid_dataset_1b = xr.Dataset(
    attrs=global_attrs,
    data_vars=dict(
        chl=xr.DataArray([1, 2, 3], dims="time", attrs=common_attrs),
        chl_unc=xr.DataArray(0, attrs=dict(units="...")),
    ),
    coords=dict(time=time_coord),
)
valid_dataset_2 = xr.Dataset(
    attrs=global_attrs,
    data_vars=dict(chl=xr.DataArray([1, 2, 3], dims="time", attrs=dict())),
    coords=dict(time=time_coord),
)
valid_dataset_3 = xr.Dataset(
    attrs=global_attrs,
    data_vars=dict(
        chl=xr.DataArray([1, 2, 3], dims="time", attrs=dict(description="Bla!"))
    ),
    coords=dict(time=time_coord),
)

invalid_dataset_0 = xr.Dataset()
invalid_dataset_1 = xr.Dataset(
    attrs=dict(),
    data_vars=dict(chl=xr.DataArray([1, 2, 3], dims="time", attrs=dict())),
    coords=dict(time=time_coord),
)
invalid_dataset_2 = xr.Dataset(
    attrs=global_attrs,
    data_vars=dict(chl=xr.DataArray([1, 2, 3], dims="time", attrs=dict())),
    coords=dict(time=time_coord),
)

ContentDescTest = RuleTester.define_test(
    "content-desc",
    ContentDesc,
    valid=[
        RuleTest(dataset=valid_dataset_0, name="0"),
        RuleTest(dataset=valid_dataset_1, name="1"),
        RuleTest(dataset=valid_dataset_1a, name="1a"),
        RuleTest(
            dataset=valid_dataset_1b, name="1b", kwargs={"ignored_vars": ["chl_unc"]}
        ),
        RuleTest(dataset=valid_dataset_2, name="2", kwargs={"commons": []}),
        RuleTest(
            dataset=valid_dataset_2, name="2", kwargs={"commons": [], "skip_vars": True}
        ),
        RuleTest(
            dataset=valid_dataset_3, name="3", kwargs={"commons": ["description"]}
        ),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=2),
        RuleTest(dataset=invalid_dataset_1, expected=6),
        RuleTest(dataset=invalid_dataset_2, kwargs={"skip_vars": True}, expected=4),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import re

from xrlint.node import DatasetNode, VariableNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleExit, RuleOp
from xrlint.util.schema import schema

DEFAULT_GLOBAL_ATTRS = ["title", "history"]
DEFAULT_COMMON_ATTRS = ["institution", "source", "references", "comment"]
DEFAULT_SKIP_VARS = False
# for the case that these are data vars by accident
DEFAULT_IGNORED_VARS = ["crs", "spatial_ref"]


@plugin.define_rule(
    "content-desc",
    version="1.0.0",
    type="suggestion",
    description=(
        "A dataset should provide information about where the data came"
        " from and what has been done to it."
        " This information is mainly for the benefit of human readers."
        " The rule accepts the following configuration parameters:\n\n"
        "- `globals`: list of names of required global attributes."
        f" Defaults to `{DEFAULT_GLOBAL_ATTRS}`.\n"
        "- `commons`: list of names of required variable attributes"
        " that can also be defined globally."
        f" Defaults to `{DEFAULT_COMMON_ATTRS}`.\n"
        "- `no_vars`: do not check variables at all."
        f" Defaults to `{DEFAULT_SKIP_VARS}`.\n"
        "- `ignored_vars`: list of ignored variables (regex patterns)."
        f" Defaults to `{DEFAULT_IGNORED_VARS}`.\n"
        ""
    ),
    docs_url=(
        "https://cfconventions.org/cf-conventions/cf-conventions.html"
        "#description-of-file-contents"
    ),
    schema=schema(
        "object",
        properties={
            "globals": schema(
                "array",
                items=schema("string"),
                default=DEFAULT_GLOBAL_ATTRS,
                title="Global attribute names",
            ),
            "commons": schema(
                "array",
                items=schema("string"),
                default=DEFAULT_COMMON_ATTRS,
                title="Common attribute names",
            ),
            "skip_vars": schema(
                "boolean",
                default=DEFAULT_SKIP_VARS,
                title="Do not check variables",
            ),
            "ignored_vars": schema(
                "array",
                items=schema("string"),
                default=DEFAULT_IGNORED_VARS,
                title="Ignored variables (regex name patterns)",
            ),
        },
    ),
)
class ContentDesc(RuleOp):
    def __init__(self, **params):
        self.global_attrs = params.get("globals", DEFAULT_GLOBAL_ATTRS)
        self.common_attrs = params.get("commons", DEFAULT_COMMON_ATTRS)
        self.skip_vars = params.get("skip_vars", DEFAULT_SKIP_VARS)
        self.ignored_vars = [
            re.compile(p) for p in params.get("ignored_vars", DEFAULT_IGNORED_VARS)
        ]

    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        dataset_attrs = node.dataset.attrs
        attr_names = (
            self.global_attrs + self.common_attrs
            if self.skip_vars
            else self.global_attrs
        )
        for attr_name in attr_names:
            if attr_name not in dataset_attrs:
                ctx.report(f"Missing attribute {attr_name!r}.")

    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        if self.skip_vars:
            # Since dataset() has already been processed,
            # no need to check other nodes.
            raise RuleExit

        if node.name not in ctx.dataset.data_vars:
            # Not a data variable
            return

        for m in self.ignored_vars:
            if m.match(str(node.name)):
                # Ignored variable
                return

        var_attrs = node.array.attrs
        dataset_attrs = ctx.dataset.attrs
        for attr_name in self.common_attrs:
            if attr_name not in var_attrs and attr_name not in dataset_attrs:
                ctx.report(f"Missing attribute {attr_name!r}.")



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import time
from typing import Any

import xarray as xr

from xrlint.config import ConfigObject
from xrlint.result import Message, Result

from ..constants import DATASET_ROOT_NAME
from .apply import apply_rule
from .rulectx import RuleContextImpl


def validate_dataset(config_obj: ConfigObject, dataset: Any, file_path: str):
    assert isinstance(config_obj, ConfigObject)
    assert dataset is not None
    assert isinstance(file_path, str)
    if isinstance(dataset, (xr.Dataset, xr.DataTree)):
        messages = _validate_dataset(config_obj, dataset, file_path, None, None)
    else:
        messages = _open_and_validate_dataset(config_obj, dataset, file_path)
    return Result(file_path=file_path, config_object=config_obj, messages=messages)


def _validate_dataset(
    config_obj: ConfigObject,
    dataset: xr.Dataset | xr.DataTree,
    file_path: str,
    file_index: int | None,
    access_latency: float | None,
) -> list[Message]:
    assert isinstance(config_obj, ConfigObject)
    assert isinstance(dataset, (xr.Dataset, xr.DataTree))
    assert isinstance(file_path, str)

    context = RuleContextImpl(
        config_obj, dataset, file_path, file_index, access_latency
    )
    for rule_id, rule_config in config_obj.rules.items():
        with context.use_state(rule_id=rule_id):
            apply_rule(context, rule_id, rule_config)
    return context.messages


def _open_and_validate_dataset(
    config_obj: ConfigObject, ds_source: Any, file_path: str
) -> list[Message]:
    assert isinstance(config_obj, ConfigObject)
    assert ds_source is not None
    assert isinstance(file_path, str)

    opener_options = config_obj.opener_options or {}
    if config_obj.processor is not None:
        processor_op = config_obj.get_processor_op(config_obj.processor)
        t0 = time.time()
        try:
            ds_path_list = processor_op.preprocess(file_path, opener_options)
        except (OSError, ValueError, TypeError) as e:
            return [new_fatal_message(str(e))]
        access_latency = time.time() - t0
        messages = [
            _validate_dataset(config_obj, ds, path, i, access_latency)
            for i, (ds, path) in enumerate(ds_path_list)
        ]
        return processor_op.postprocess(messages, file_path)
    else:
        try:
            dataset, access_latency = _open_dataset(
                ds_source, opener_options, file_path
            )
        except (OSError, ValueError, TypeError) as e:
            return [new_fatal_message(str(e))]
        with dataset:
            return _validate_dataset(
                config_obj, dataset, file_path, None, access_latency
            )


def _open_dataset(
    ds_source: Any, opener_options: dict[str, Any] | None, file_path: str
) -> tuple[xr.Dataset | xr.DataTree, float]:
    """Open a dataset."""
    engine = opener_options.pop("engine", None)
    if engine is None and (file_path.endswith(".zarr") or file_path.endswith(".zarr/")):
        engine = "zarr"
    try:
        t0 = time.time()
        result = xr.open_datatree(ds_source, engine=engine, **(opener_options or {}))
        # When opening no-group Zarr datasets we get with xarray 2025.1.2:
        #
        #   File "<...>/site-packages/xarray/backends/zarr.py", line 741, in __init__
        #     self._read_only = self.zarr_group.read_only
        #                       ^^^^^^^^^^^^^^^^^^^^^^^^^
        # AttributeError: 'NoneType' object has no attribute 'read_only'
    except (OSError, ValueError, TypeError, AttributeError):
        t0 = time.time()
        result = xr.open_dataset(ds_source, engine=engine, **(opener_options or {}))
    return result, time.time() - t0


def new_fatal_message(message: str) -> Message:
    return Message(
        message=message,
        fatal=True,
        severity=2,
        node_path=DATASET_ROOT_NAME,
    )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.node import XarrayNode


class XarrayNodeTest(TestCase):
    def node(self, path: str):
        return XarrayNode(path=path, parent=None)

    def attrs_node(self):
        return self.node("dataset.attrs")

    def coords_node(self):
        return self.node("dataset.coords['x']")

    def data_var_node(self):
        return self.node("dataset.data_vars['v']")

    def test_in_coords(self):
        self.assertEqual(False, self.attrs_node().in_coords())
        self.assertEqual(True, self.coords_node().in_coords())
        self.assertEqual(False, self.data_var_node().in_coords())

    def test_in_data_vars(self):
        self.assertEqual(False, self.attrs_node().in_data_vars())
        self.assertEqual(False, self.coords_node().in_data_vars())
        self.assertEqual(True, self.data_var_node().in_data_vars())

    def test_in_root(self):
        self.assertEqual(True, self.attrs_node().in_root())
        self.assertEqual(False, self.coords_node().in_root())
        self.assertEqual(False, self.data_var_node().in_root())



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import DatasetNode
from xrlint.plugins.xcube.constants import LAT_NAME, LON_NAME
from xrlint.plugins.xcube.plugin import plugin
from xrlint.rule import RuleContext, RuleOp

INVALID_LAT_NAMES = {"ltd", "latitude"}
INVALID_LON_NAMES = {"lng", "long", "longitude"}


@plugin.define_rule(
    "lat-lon-naming",
    version="1.0.0",
    type="problem",
    description=(
        f"Latitude and longitude coordinates and dimensions"
        f" should be called {LAT_NAME!r} and {LON_NAME!r}."
    ),
    docs_url="https://xcube.readthedocs.io/en/latest/cubespec.html#spatial-reference",
)
class LatLonNaming(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        lon_ok = _check(
            ctx, "variable", node.dataset.variables.keys(), INVALID_LON_NAMES, LON_NAME
        )
        lat_ok = _check(
            ctx, "variable", node.dataset.variables.keys(), INVALID_LAT_NAMES, LAT_NAME
        )
        if lon_ok and lat_ok:
            # If variables have been reported,
            # we should not need to report (their) coordinates
            _check(
                ctx,
                "dimension",
                node.dataset.sizes.keys(),
                INVALID_LON_NAMES,
                LON_NAME,
            )
            _check(
                ctx,
                "dimension",
                node.dataset.sizes.keys(),
                INVALID_LAT_NAMES,
                LAT_NAME,
            )


def _check(ctx, names_name, names, invalid_names, valid_name):
    names = [str(n) for n in names]  # xarray keys are Hashable, not str
    found_names = [
        n
        for n in names
        if (n.lower() in invalid_names) or (n.lower() == valid_name and n != valid_name)
    ]
    if found_names:
        ctx.report(
            f"The {names_name} {found_names[0]!r} should be named {valid_name!r}.",
            suggestions=[f"Rename {names_name} to {valid_name!r}."],
        )
        return False
    else:
        return True



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np
import xarray as xr

from xrlint.plugins.core.rules.lat_lon_coordinate import LatCoordinate, LonCoordinate
from xrlint.testing import RuleTest, RuleTester

valid_dataset_0 = xr.Dataset()
valid_dataset_1 = xr.Dataset(
    coords={
        "lat": xr.DataArray(
            np.array([3, 4, 5]),
            dims="lat",
            attrs={
                "units": "degrees_north",
                "standard_name": "latitude",
                "long_name": "latitude",
            },
        ),
        "lon": xr.DataArray(
            np.array([-2, -1, 0, 1]),
            dims="lon",
            attrs={
                "units": "degrees_east",
                "standard_name": "longitude",
                "long_name": "longitude",
            },
        ),
    },
    data_vars={
        "mask": xr.DataArray(
            [[10, 20, 30, 40], [30, 40, 50, 60], [50, 60, 70, 80]], dims=("lat", "lon")
        )
    },
)

# Valid, because the coord names doesn't matter as long their metadata is ok
valid_dataset_2 = valid_dataset_1.rename_vars({"lon": "x", "lat": "y"})

# Valid, because the coord units have aliases
valid_dataset_3 = valid_dataset_1.copy()
valid_dataset_3.lat.attrs["units"] = "degreeN"
valid_dataset_3.lon.attrs["units"] = "degreeE"

# Valid, because the coord units have aliases
valid_dataset_4 = valid_dataset_1.copy()
del valid_dataset_4.lat.attrs["standard_name"]
del valid_dataset_4.lon.attrs["standard_name"]
valid_dataset_4.lat.attrs["axis"] = "Y"
valid_dataset_4.lon.attrs["axis"] = "X"

invalid_lat_dataset_0 = valid_dataset_1.copy()
del invalid_lat_dataset_0.lat.attrs["standard_name"]

invalid_lon_dataset_0 = valid_dataset_1.copy()
del invalid_lon_dataset_0.lon.attrs["units"]

invalid_lat_dataset_1 = valid_dataset_1.copy()
invalid_lat_dataset_1.lat.attrs["units"] = "deg"

invalid_lon_dataset_1 = valid_dataset_1.copy()
invalid_lon_dataset_1.lon.attrs["long_name"] = "poo"

invalid_lat_dataset_2 = valid_dataset_1.copy()
del invalid_lat_dataset_2.lat.attrs["standard_name"]
invalid_lat_dataset_2.lat.attrs["axis"] = "y"  # should be "Y"

invalid_lon_dataset_2 = valid_dataset_1.copy()
del invalid_lon_dataset_2.lon.attrs["standard_name"]
invalid_lon_dataset_2.lon.attrs["axis"] = "x"  # should be "X"

LatCoordsTest = RuleTester.define_test(
    "lat-coordinate",
    LatCoordinate,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
        RuleTest(dataset=valid_dataset_3),
        RuleTest(dataset=valid_dataset_4),
        RuleTest(dataset=invalid_lon_dataset_0),
        RuleTest(dataset=invalid_lon_dataset_1),
        RuleTest(dataset=invalid_lon_dataset_2),
    ],
    invalid=[
        RuleTest(dataset=invalid_lat_dataset_0, expected=1),
        RuleTest(dataset=invalid_lat_dataset_1, expected=1),
        RuleTest(dataset=invalid_lat_dataset_2, expected=2),
    ],
)

LonCoordsTest = RuleTester.define_test(
    "lon-coordinate",
    LonCoordinate,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
        RuleTest(dataset=valid_dataset_3),
        RuleTest(dataset=valid_dataset_4),
        RuleTest(dataset=invalid_lat_dataset_0),
        RuleTest(dataset=invalid_lat_dataset_1),
        RuleTest(dataset=invalid_lat_dataset_2),
    ],
    invalid=[
        RuleTest(dataset=invalid_lon_dataset_0, expected=1),
        RuleTest(dataset=invalid_lon_dataset_1, expected=1),
        RuleTest(dataset=invalid_lon_dataset_2, expected=2),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.constants import CORE_DOCS_URL, CORE_PLUGIN_NAME
from xrlint.plugin import new_plugin
from xrlint.version import version

plugin = new_plugin(
    name=CORE_PLUGIN_NAME,
    version=version,
    ref="xrlint.plugins.core:export_plugin",
    docs_url=CORE_DOCS_URL,
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import os
import shutil
import tempfile
from unittest import TestCase

import click.testing
import xarray as xr
from click.testing import CliRunner

from xrlint.cli.constants import DEFAULT_CONFIG_FILE_YAML
from xrlint.cli.main import main
from xrlint.version import version

from .helpers import text_file

no_match_config_yaml = """
- ignores: ['**/*.nc', '**/*.zarr']
  rules:
    dataset-title-attr: error
"""


# noinspection PyTypeChecker
class CliMainTest(TestCase):
    files = ["dataset1.zarr", "dataset1.nc", "dataset2.zarr", "dataset2.nc"]

    ok_config_yaml = "- rules:\n    var-units: error\n"
    fail_config_yaml = "- rules:\n    conventions: error\n"
    # noinspection SpellCheckingInspection
    invalid_config_yaml = "- recommentet\n"

    datasets = dict(
        dataset1=xr.Dataset(attrs={"title": "Test 1"}),
        dataset2=xr.Dataset(
            attrs={"title": "Test 2"},
            data_vars={"v": xr.DataArray([1, 2, 3], attrs={"units": "m/s"})},
        ),
    )

    temp_dir: str
    last_cwd: str

    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.mkdtemp(prefix="xrlint-")
        cls.last_cwd = os.getcwd()
        os.chdir(cls.temp_dir)

        for file in cls.files:
            name, ext = file.split(".")
            if ext == "zarr":
                cls.datasets[name].to_zarr(file)
            else:
                cls.datasets[name].to_netcdf(file)

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.last_cwd)
        shutil.rmtree(cls.temp_dir)

    def xrlint(self, *args: tuple[str, ...]) -> click.testing.Result:
        runner = CliRunner()
        result = runner.invoke(main, args)
        if not isinstance(result.exception, SystemExit):
            import traceback

            traceback.print_exception(result.exception)
            self.assertIsNone(result.exception)
        return result

    def test_no_files_no_config(self):
        result = self.xrlint()
        self.assertEqual("", result.output)
        self.assertEqual(0, result.exit_code)

    def test_config_no_files(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, self.ok_config_yaml):
            result = self.xrlint()
            self.assertEqual("", result.output)
            self.assertEqual(0, result.exit_code)

    def test_files_no_config(self):
        result = self.xrlint(*self.files)
        self.assertIn("Warning: no configuration file found.\n", result.output)
        self.assertEqual(1, result.exit_code)

    def test_files_no_config_lookup(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, self.ok_config_yaml):
            result = self.xrlint("--no-config-lookup", *self.files)
            self.assertEqual("Error: no rules configured\n", result.output)
            self.assertEqual(1, result.exit_code)

    def test_files_one_rule(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, self.ok_config_yaml):
            result = self.xrlint("--no-color", *self.files)
            self.assertEqual(
                "\n"
                "dataset1.zarr - ok\n\n"
                "dataset1.nc - ok\n\n"
                "dataset2.zarr - ok\n\n"
                "dataset2.nc - ok\n\n"
                "no problems\n\n",
                result.output,
            )
            self.assertEqual(0, result.exit_code)

        with text_file(DEFAULT_CONFIG_FILE_YAML, self.fail_config_yaml):
            result = self.xrlint(*self.files)
            self.assertIn("Missing attribute 'Conventions'.", result.output)
            self.assertEqual(1, result.exit_code)

    def test_dir_one_rule(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, self.ok_config_yaml):
            result = self.xrlint("--no-color", ".")
            prefix = self.temp_dir.replace("\\", "/")
            self.assertIn(f"{prefix}/dataset1.zarr - ok\n\n", result.output)
            self.assertIn(f"{prefix}/dataset1.nc - ok\n\n", result.output)
            self.assertIn(f"{prefix}/dataset2.zarr - ok\n\n", result.output)
            self.assertIn(f"{prefix}/dataset2.nc - ok\n\n", result.output)
            self.assertIn("no problems\n\n", result.output)
            self.assertEqual(0, result.exit_code)

        with text_file(DEFAULT_CONFIG_FILE_YAML, self.fail_config_yaml):
            result = self.xrlint(*self.files)
            self.assertIn("Missing attribute 'Conventions'.", result.output)
            self.assertEqual(1, result.exit_code)

    def test_color_no_color(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, self.ok_config_yaml):
            result = self.xrlint("--no-color", *self.files)
            self.assertIsNone(result.exception)
            self.assertEqual(
                "\n"
                "dataset1.zarr - ok\n\n"
                "dataset1.nc - ok\n\n"
                "dataset2.zarr - ok\n\n"
                "dataset2.nc - ok\n\n"
                "no problems\n\n",
                result.output,
            )
            self.assertEqual(0, result.exit_code)

        with text_file(DEFAULT_CONFIG_FILE_YAML, self.ok_config_yaml):
            runner = CliRunner()
            result = runner.invoke(main, ["--color"] + self.files)
            self.assertEqual(
                "\n\x1b[4mdataset1.zarr\x1b[0m - ok\n\n"
                "\x1b[4mdataset1.nc\x1b[0m - ok\n\n"
                "\x1b[4mdataset2.zarr\x1b[0m - ok\n\n"
                "\x1b[4mdataset2.nc\x1b[0m - ok\n\n"
                "no problems\n\n",
                result.output,
            )
            self.assertEqual(0, result.exit_code)

    def test_files_with_invalid_config(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, self.invalid_config_yaml):
            result = self.xrlint("--no-color", *self.files)
            # noinspection SpellCheckingInspection
            self.assertEqual(
                f"Error: {DEFAULT_CONFIG_FILE_YAML}:"
                " configuration 'recommentet' not found\n",
                result.output,
            )
            self.assertEqual(1, result.exit_code)

    def test_files_with_rule_option(self):
        result = self.xrlint("--rule", "conventions: error", *self.files)
        self.assertIn("Missing attribute 'Conventions'.", result.output)
        self.assertEqual(1, result.exit_code)

    def test_files_with_max_warnings(self):
        result = self.xrlint(
            "--rule", "conventions: warn", "--max-warnings", "0", *self.files
        )
        self.assertIn("Maximum number of warnings exceeded.", result.output)
        self.assertEqual(1, result.exit_code)

    def test_files_with_plugin_and_rule_options(self):
        result = self.xrlint(
            "--plugin",
            "xrlint.plugins.xcube",
            "--rule",
            "xcube/any-spatial-data-var: error",
            *self.files,
        )
        self.assertIn("No spatial data variables found.", result.output)
        self.assertIn("xcube/any-spatial-data-var", result.output)
        self.assertEqual(1, result.exit_code)

    def test_files_with_output_file(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, self.ok_config_yaml):
            result = self.xrlint("-o", "memory://report.txt", *self.files)
            self.assertEqual("", result.output)
            self.assertEqual(0, result.exit_code)

    def test_files_but_config_file_missing(self):
        result = self.xrlint("-c", "pippo.py", *self.files)
        self.assertIn("Error: file not found: pippo.py", result.output)
        self.assertEqual(1, result.exit_code)

    def test_files_with_format_json(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, self.ok_config_yaml):
            result = self.xrlint("-f", "json", *self.files)
            self.assertIn('"results": [\n', result.output)
            self.assertEqual(0, result.exit_code)

    def test_files_with_format_html(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, self.ok_config_yaml):
            result = self.xrlint("-f", "html", *self.files)
            self.assertIn("<h3>Results</h3>", result.output)
            self.assertEqual(0, result.exit_code)

    def test_file_does_not_match(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, no_match_config_yaml):
            result = self.xrlint("test.zarr")
            self.assertIn(
                "No configuration given or matches 'test.zarr'.", result.output
            )
            self.assertEqual(1, result.exit_code)

    def test_print_config_option(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, self.ok_config_yaml):
            result = self.xrlint("--print-config", "dataset2.zarr")
            self.assertEqual(
                (
                    "{\n"
                    '  "plugins": {\n'
                    '    "__core__": "xrlint.plugins.core:export_plugin"\n'
                    "  },\n"
                    '  "rules": {\n'
                    '    "var-units": 2\n'
                    "  }\n"
                    "}\n"
                ),
                result.output,
            )
            self.assertEqual(0, result.exit_code)

    def test_files_with_invalid_format_option(self):
        with text_file(DEFAULT_CONFIG_FILE_YAML, self.ok_config_yaml):
            result = self.xrlint("-f", "foo", *self.files)
            self.assertIn(
                "Error: unknown format 'foo'. The available formats are '",
                result.output,
            )
            self.assertEqual(1, result.exit_code)

    def test_init(self):
        config_file = DEFAULT_CONFIG_FILE_YAML
        exists = os.path.exists(config_file)
        self.assertFalse(exists)
        try:
            result = self.xrlint("--init")
            self.assertEqual(
                f"Configuration template written to {config_file}\n", result.output
            )
            self.assertEqual(result.exit_code, 0)
            exists = os.path.exists(config_file)
            self.assertTrue(exists)
        finally:
            if exists:
                os.remove(config_file)

    def test_init_exists(self):
        config_file = DEFAULT_CONFIG_FILE_YAML
        exists = os.path.exists(config_file)
        self.assertFalse(exists)
        with text_file(config_file, ""):
            result = self.xrlint("--init")
            self.assertEqual(
                f"Error: file {config_file} already exists.\n", result.output
            )
            self.assertEqual(result.exit_code, 1)


# noinspection PyTypeChecker
class CliMainMetaTest(TestCase):
    def test_help(self):
        runner = CliRunner()
        result = runner.invoke(main, ["--help"])
        self.assertIn("Usage: xrlint [OPTIONS] [FILES]...\n", result.output)
        self.assertEqual(result.exit_code, 0)

    def test_version(self):
        runner = CliRunner()
        result = runner.invoke(main, ["--version"])
        self.assertIn(f"xrlint, version {version}", result.output)
        self.assertEqual(result.exit_code, 0)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Type

import xarray as xr

from xrlint.operation import Operation, OperationMeta
from xrlint.result import Message


class ProcessorOp(ABC):
    """Implements the processor operations."""

    @abstractmethod
    def preprocess(
        self, file_path: str, opener_options: dict[str, Any]
    ) -> list[tuple[xr.Dataset | xr.DataTree, str]]:
        """Pre-process a dataset given by its `file_path` and `opener_options`.
        In this method you use the `file_path` to read zero, one, or more
        datasets to lint.

        Args:
            file_path: A file path
            opener_options: The configuration's `opener_options`.

        Returns:
            A list of (dataset or datatree, file_path) pairs
        """

    @abstractmethod
    def postprocess(
        self, messages: list[list[Message]], file_path: str
    ) -> list[Message]:
        """Post-process the outputs of each dataset from `preprocess()`.

        Args:
            messages: contains two-dimensional array of ´Message´ objects
                where each top-level array item contains array of lint messages
                related to the dataset that was returned in array from
                `preprocess()` method
            file_path: The corresponding file path

        Returns:
            A one-dimensional array (list) of the messages you want to keep
        """


@dataclass(kw_only=True)
class ProcessorMeta(OperationMeta):
    """Processor metadata."""

    name: str
    """Processor name."""

    version: str = "0.0.0"
    """Processor version."""

    """Processor description. Optional."""
    description: str | None = None

    ref: str | None = None
    """Processor reference.
    Specifies the location from where the processor can be
    dynamically imported.
    Must have the form "<module>:<attr>", if given.
    """

    @classmethod
    def value_name(cls) -> str:
        return "processor_meta"

    @classmethod
    def value_type_name(cls) -> str:
        return "ProcessorMeta | dict"


@dataclass(frozen=True, kw_only=True)
class Processor(Operation):
    """Processors tell XRLint how to process files other than
    standard xarray datasets.
    """

    meta: ProcessorMeta
    """Information about the processor."""

    op_class: Type[ProcessorOp]
    """A class that implements the processor operations."""

    # Not yet:
    # supports_auto_fix: bool = False
    # """`True` if this processor supports auto-fixing of datasets."""

    @classmethod
    def meta_class(cls) -> Type:
        return ProcessorMeta

    @classmethod
    def op_base_class(cls) -> Type:
        return ProcessorOp

    @classmethod
    def value_name(cls) -> str:
        return "processor"


def define_processor(
    name: str | None = None,
    version: str = "0.0.0",
    registry: dict[str, Processor] | None = None,
    op_class: Type[ProcessorOp] | None = None,
) -> Callable[[Any], Type[ProcessorOp]] | Processor:
    """Define a processor.

    This function can be used to decorate your processor operation class
    definitions. When used as a decorator, the decorated operator class
    will receive a `meta` attribute of type
    [ProcessorMeta][xrlint.processor.ProcessorMeta].
    In addition, the `registry` if given, will be updated using `name`
    as key and a new [Processor][xrlint.processor.Processor] as value.

    Args:
        name: Processor name,
            see [ProcessorMeta][xrlint.processor.ProcessorMeta].
        version: Processor version,
            see [ProcessorMeta][xrlint.processor.ProcessorMeta].
        registry: Processor registry. Can be provided to register the
            defined processor using its `name`.
        op_class: Processor operation class. Must be `None`
            if this function is used as a class decorator.

    Returns:
        A decorator function, if `op_class` is `None` otherwise
            the value of `op_class`.

    Raises:
        TypeError: If either `op_class` or the decorated object is not a
            a class derived from [ProcessorOp][xrlint.processor.ProcessorOp].
    """
    return Processor.define_operation(
        op_class, registry=registry, meta_kwargs=dict(name=name, version=version)
    )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from dataclasses import dataclass

from xrlint.util.filepattern import FilePattern


@dataclass(frozen=True)
class FileFilter:
    """Encapsulates the file filtering mechanism using `files` and `ignores`.

    Args:
        files: File path patterns for files to be included.
        ignores: File path patterns for files to be excluded.
    """

    files: tuple[FilePattern, ...] = ()
    ignores: tuple[FilePattern, ...] = ()

    @classmethod
    def from_patterns(
        cls, files: list[str] | None, ignores: list[str] | None
    ) -> "FileFilter":
        return FileFilter(
            cls.patterns_to_matchers(files),
            cls.patterns_to_matchers(ignores, flip_negate=True),
        )

    @classmethod
    def patterns_to_matchers(
        cls, patterns: list[str] | None, flip_negate: bool = False
    ) -> tuple[FilePattern, ...]:
        matchers = (FilePattern(p, flip_negate=flip_negate) for p in patterns or ())
        return tuple(m for m in matchers if not (m.empty or m.comment))

    @property
    def empty(self) -> bool:
        return not (self.files or self.ignores)

    def merge(self, file_filter: "FileFilter") -> "FileFilter":
        return FileFilter(
            self.files + file_filter.files,  # note, we should exclude duplicates
            self.ignores + file_filter.ignores,
        )

    def accept(self, file_path) -> bool:
        if self.files:
            included = False
            for p in self.files:
                if p.match(file_path):
                    included = True
                    break
            if not included:
                return False

        excluded = False
        for p in self.ignores:
            if not p.negate:
                if excluded:
                    # Already excluded, no need to check further
                    return False
                excluded = p.match(file_path)
            else:
                if excluded and p.match(file_path):
                    # Negate the old excluded status on match
                    excluded = False

        return not excluded



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import math

from xrlint.node import DatasetNode
from xrlint.plugins.xcube.plugin import plugin
from xrlint.plugins.xcube.util import get_dataset_level_info, get_spatial_size
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "ml-dataset-xy",
    version="1.0.0",
    type="problem",
    description=(
        "Multi-level dataset levels should provide spatial resolutions"
        " decreasing by powers of two."
    ),
    docs_url="https://xcube.readthedocs.io/en/latest/mldatasets.html#definition",
)
class MLDatasetXY(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        level_info = get_dataset_level_info(node.dataset)
        if level_info is None:
            # ok, this rules applies only to level datasets opened
            # by the xcube multi-level processor
            return

        level = level_info.level
        if level == 0:
            # ok, this rule does only apply to level > 0
            return

        datasets = level_info.datasets
        level_0_dataset, _ = datasets[0]
        l0_size = get_spatial_size(level_0_dataset)
        if l0_size is None:
            # ok, maybe no spatial data vars?
            return

        (x_name, level_0_width), (y_name, level_0_height) = l0_size
        level_width = node.dataset.sizes.get(x_name)
        level_height = node.dataset.sizes.get(y_name)
        expected_level_width = math.ceil(level_0_width >> level)
        expected_level_height = math.ceil(level_0_height >> level)

        if level_width != expected_level_width:
            ctx.report(
                f"Expected size of dimension {x_name!r} in level {level}"
                f" to be {expected_level_width}, but was {level_width}."
            )

        if level_height != expected_level_height:
            ctx.report(
                f"Expected size of dimension {y_name!r} in level {level}"
                f" to be {expected_level_height}, but was {level_height}."
            )

        # Here: check spatial coordinates...



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np
import xarray as xr

from xrlint.plugins.xcube.rules.lat_lon_naming import LatLonNaming
from xrlint.testing import RuleTest, RuleTester


def make_dataset(lat_dim: str, lon_dim: str):
    dims = ["time", lat_dim, lon_dim]
    n = 3
    return xr.Dataset(
        attrs=dict(title="v-data"),
        coords={
            lon_dim: xr.DataArray(
                np.linspace(0, 1, n), dims=lon_dim, attrs={"units": "m"}
            ),
            lat_dim: xr.DataArray(
                np.linspace(0, 1, n), dims=lat_dim, attrs={"units": "m"}
            ),
            "time": xr.DataArray(
                list(range(2010, 2010 + n)), dims="time", attrs={"units": "years"}
            ),
        },
        data_vars={
            "chl": xr.DataArray(
                np.random.random((n, n, n)), dims=dims, attrs={"units": "mg/m^-3"}
            ),
            "tsm": xr.DataArray(
                np.random.random((n, n, n)), dims=dims, attrs={"units": "mg/m^-3"}
            ),
            "avg_temp": xr.DataArray(
                np.random.random(n), dims=dims[0], attrs={"units": "kelvin"}
            ),
            "mask": xr.DataArray(np.random.random((n, n)), dims=dims[-2:]),
        },
    )


valid_dataset_1 = make_dataset("lat", "lon")

invalid_dataset_1 = make_dataset("lat", "long")
invalid_dataset_2 = make_dataset("lat", "longitude")
invalid_dataset_3 = make_dataset("lat", "Lon")

invalid_dataset_4 = make_dataset("ltd", "lon")
invalid_dataset_5 = make_dataset("latitude", "lon")
invalid_dataset_6 = make_dataset("Lat", "lon")

LatLonNamingTest = RuleTester.define_test(
    "lat-lon-naming",
    LatLonNaming,
    valid=[
        RuleTest(dataset=valid_dataset_1),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_1, expected=1),
        RuleTest(dataset=invalid_dataset_2, expected=1),
        RuleTest(dataset=invalid_dataset_3, expected=1),
        RuleTest(dataset=invalid_dataset_4, expected=1),
        RuleTest(dataset=invalid_dataset_5, expected=1),
        RuleTest(dataset=invalid_dataset_6, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

import pytest

from xrlint.util.schema import schema


class SchemaTest(TestCase):
    def test_type_name(self):
        self.assertEqual({}, schema())
        self.assertEqual({}, schema(type=None))
        self.assertEqual({"type": "null"}, schema("null"))
        self.assertEqual({"type": "boolean"}, schema("boolean"))
        self.assertEqual({"type": "integer"}, schema("integer"))
        self.assertEqual({"type": "number"}, schema("number"))
        self.assertEqual({"type": "string"}, schema("string"))
        self.assertEqual({"type": "object"}, schema("object"))
        self.assertEqual({"type": "array"}, schema("array"))

    def test_type_name_list(self):
        self.assertEqual({}, schema())
        self.assertEqual({}, schema([]))
        self.assertEqual({"type": ["array", "string"]}, schema(["array", "string"]))

    # noinspection PyTypeChecker,PyMethodMayBeStatic
    def test_type_name_invalid(self):
        with pytest.raises(
            TypeError, match="type must be of type str|list[str], but got str"
        ):
            schema(type=str)
        with pytest.raises(
            ValueError,
            match=(
                "type name must be one of "
                "'null', 'boolean', 'integer', 'number',"
                " 'string', 'array', 'object',"
                " but got 'float'"
            ),
        ):
            schema(type="float")

    # noinspection PyTypeChecker,PyMethodMayBeStatic
    def test_type_name_list_invalid(self):
        with pytest.raises(
            TypeError, match="type must be of type str|list[str], but got int"
        ):
            schema(type=["string", 2])
        with pytest.raises(
            ValueError,
            match=(
                "type name must be one of"
                " 'null', 'boolean', 'integer', 'number',"
                " 'string', 'array', 'object',"
                " but got 'list'"
            ),
        ):
            schema(type=["string", "list"])

    def test_properties(self):
        self.assertEqual({}, schema(properties=None))
        self.assertEqual({"properties": {}}, schema(properties={}))
        self.assertEqual(
            {"properties": {"a": {"type": "string"}, "b": {}}},
            schema(properties={"a": schema("string"), "b": schema()}),
        )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Final

from xrlint.node import DatasetNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleExit, RuleOp
from xrlint.util.formatting import format_count
from xrlint.util.schema import schema

DEFAULT_THRESHOLD: Final = 2.5  # seconds


@plugin.define_rule(
    "access-latency",
    version="1.0.0",
    description=(
        "Ensure that the time it takes to open a dataset from its source"
        " does a exceed a given `threshold` in seconds."
        f" The default threshold is `{DEFAULT_THRESHOLD}`."
    ),
    schema=schema(
        "object",
        properties={
            "threshold": schema(
                "number",
                exclusiveMinimum=0,
                default=DEFAULT_THRESHOLD,
                title="Threshold time in seconds",
            )
        },
    ),
)
class AccessLatency(RuleOp):
    def __init__(self, threshold: float = DEFAULT_THRESHOLD):
        self.threshold = threshold

    def validate_dataset(self, ctx: RuleContext, node: DatasetNode) -> None:
        if ctx.access_latency is not None and ctx.access_latency > self.threshold:
            ctx.report(
                f"Access latency exceeds threshold: {ctx.access_latency:.1f}"
                f" > {format_count(self.threshold, 'second')}."
            )
        raise RuleExit



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Any

import xarray as xr

from xrlint.node import VariableNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleOp

LAT_NAME = "latitude"
LON_NAME = "longitude"

LAT_ALIASES = {LAT_NAME, "lat"}
LON_ALIASES = {LON_NAME, "lon", "long"}

LAT_UNITS = "degrees_north"
LON_UNITS = "degrees_east"

LAT_UNITS_ALIASES = {"degree_north", "degree_N", "degrees_N", "degreeN", "degreesN"}
LON_UNITS_ALIASES = {"degree_east", "degree_E", "degrees_E", "degreeE", "degreesE"}


@plugin.define_rule(
    "lat-coordinate",
    version="1.0.0",
    type="problem",
    description="Latitude coordinate should have standard units and standard names.",
    docs_url=(
        "https://cfconventions.org/cf-conventions/cf-conventions.html"
        "#latitude-coordinate"
    ),
)
class LatCoordinate(RuleOp):
    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        if node.name in ctx.dataset.coords and _is_lat_var(str(node.name), node.array):
            _maybe_report(
                ctx,
                node.array.attrs,
                LAT_UNITS,
                LAT_UNITS_ALIASES,
                LAT_NAME,
                "Y",
            )


@plugin.define_rule(
    "lon-coordinate",
    version="1.0.0",
    type="problem",
    description="Longitude coordinate should have standard units and standard names.",
    docs_url=(
        "https://cfconventions.org/cf-conventions/cf-conventions.html"
        "#longitude-coordinate"
    ),
)
class LonCoordinate(RuleOp):
    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        if node.name in ctx.dataset.coords and _is_lon_var(str(node.name), node.array):
            _maybe_report(
                ctx,
                node.array.attrs,
                LON_UNITS,
                LON_UNITS_ALIASES,
                LON_NAME,
                "X",
            )


def _is_lat_var(var_name: str, var: xr.DataArray) -> bool:
    return _is_var(var_name.lower(), var, LAT_ALIASES)


def _is_lon_var(var_name: str, var: xr.DataArray) -> bool:
    return _is_var(var_name.lower(), var, LON_ALIASES)


def _is_var(var_name: str, var: xr.DataArray, name_aliases: set[str]) -> bool:
    return bool(
        name_aliases.intersection(
            {
                var_name,
                var.attrs.get("standard_name"),
                var.attrs.get("long_name"),
            }
        )
    )


def _maybe_report(
    ctx: RuleContext,
    attrs: dict[str, Any],
    expected_units: str,
    expected_units_aliases: set[str],
    expected_name: str,
    expected_axis: str,
):
    _maybe_report_attr(
        ctx,
        attrs,
        "units",
        expected_units,
        expected_units_aliases,
        None,
        None,
    )
    _maybe_report_attr(
        ctx,
        attrs,
        "standard_name",
        expected_name,
        None,
        "axis",
        expected_axis,
    )
    _maybe_report_attr(
        ctx,
        attrs,
        "long_name",
        expected_name,
        None,
        None,
        None,
    )


def _maybe_report_attr(
    ctx: RuleContext,
    attrs: dict[str, Any],
    attr_name: str,
    expected_value: str,
    expected_value_aliases: set[str] | None,
    alt_attr_name: str | None,
    expected_alt_value: str | None,
):
    actual_value = attrs.get(attr_name)
    if alt_attr_name:
        actual_alt_value = attrs.get(alt_attr_name)
        has_alt_value = actual_alt_value is not None
    else:
        actual_alt_value = None
        has_alt_value = False

    if not actual_value and not has_alt_value:
        ctx.report(f"Missing attribute {attr_name!r} with value {expected_value!r}.")
    else:
        value_ok = actual_value == expected_value or (
            expected_value_aliases and actual_value in expected_value_aliases
        )
        alt_value_ok = has_alt_value and actual_alt_value == expected_alt_value
        if not value_ok and not alt_value_ok:
            ctx.report(
                f"Attribute {attr_name!r} should be {expected_value!r},"
                f" was {actual_value!r}."
            )
        if has_alt_value and not alt_value_ok:
            ctx.report(
                f"Attribute {alt_attr_name!r} should be {expected_alt_value!r},"
                f" was {actual_alt_value!r}."
            )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Final

CORE_PLUGIN_NAME: Final = "__core__"
CORE_DOCS_URL = "https://bcdev.github.io/xrlint/rule-ref"

DATATREE_ROOT_NAME: Final = "dt"
DATASET_ROOT_NAME: Final = "ds"
MISSING_DATATREE_FILE_PATH: Final = "<datatree>"
MISSING_DATASET_FILE_PATH: Final = "<dataset>"

SEVERITY_ERROR: Final = 2
SEVERITY_WARN: Final = 1
SEVERITY_OFF: Final = 0

SEVERITY_NAME_TO_CODE: Final = {
    "error": SEVERITY_ERROR,
    "warn": SEVERITY_WARN,
    "off": SEVERITY_OFF,
}
SEVERITY_CODE_TO_NAME: Final = {v: k for k, v in SEVERITY_NAME_TO_CODE.items()}
SEVERITY_CODE_TO_CODE: Final = {v: v for v in SEVERITY_NAME_TO_CODE.values()}
SEVERITY_CODE_TO_COLOR = {2: "red", 1: "blue", 0: "green", None: ""}

SEVERITY_ENUM: Final[dict[int | str, int]] = (
    SEVERITY_NAME_TO_CODE | SEVERITY_CODE_TO_CODE
)
SEVERITY_ENUM_TEXT: Final = ", ".join(f"{k!r}" for k in SEVERITY_ENUM.keys())



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import VariableNode
from xrlint.plugins.xcube.constants import LAT_NAME, LON_NAME, TIME_NAME, X_NAME, Y_NAME
from xrlint.plugins.xcube.plugin import plugin
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "cube-dims-order",
    version="1.0.0",
    type="problem",
    description=(
        f"Order of dimensions in spatio-temporal datacube variables"
        f" should be [{TIME_NAME}, ..., {Y_NAME}, {X_NAME}]."
    ),
    docs_url=(
        "https://xcube.readthedocs.io/en/latest/cubespec.html#data-model-and-format"
    ),
)
class CubeDimsOrder(RuleOp):
    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        if node.in_data_vars():
            dims = list(node.array.dims)
            indexes = {d: i for i, d in enumerate(node.array.dims)}

            yx_names = None
            if X_NAME in indexes and Y_NAME in indexes:
                yx_names = [Y_NAME, X_NAME]
            elif LON_NAME in indexes and LAT_NAME in indexes:
                yx_names = [LAT_NAME, LON_NAME]
            else:
                # Note, we could get yx_names also from grid-mapping,
                # which would be more generic.
                pass
            if yx_names is None:
                # This rule only applies to spatial dimensions
                return

            t_name = None
            if TIME_NAME in indexes:
                t_name = TIME_NAME

            n = len(dims)
            t_index = indexes[t_name] if t_name else None
            y_index = indexes[yx_names[0]]
            x_index = indexes[yx_names[1]]

            yx_ok = y_index == n - 2 and x_index == n - 1
            t_ok = t_index is None or t_index == 0
            if not yx_ok or not t_ok:
                if t_index is None:
                    expected_dims = [d for d in dims if d not in yx_names] + yx_names
                else:
                    expected_dims = (
                        [t_name]
                        + [d for d in dims if d != t_name and d not in yx_names]
                        + yx_names
                    )
                # noinspection PyTypeChecker
                ctx.report(
                    f"Order of dimensions should be"
                    f" {','.join(expected_dims)}, but found {','.join(dims)}.",
                    suggestions=["Use xarray.transpose(...) to reorder dimensions."],
                )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.formatters import export_formatters


class ImportFormattersTest(TestCase):
    def test_import_formatters(self):
        registry = export_formatters()
        self.assertEqual(
            {
                "html",
                "json",
                "simple",
            },
            set(registry.keys()),
        )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Any
from unittest import TestCase

import pytest
import xarray as xr

from xrlint.plugin import new_plugin
from xrlint.processor import Processor, ProcessorMeta, ProcessorOp, define_processor
from xrlint.result import Message


class ProcessorMetaTest(TestCase):
    def test_class_props(self):
        self.assertEqual("processor_meta", ProcessorMeta.value_name())
        self.assertEqual("ProcessorMeta | dict", ProcessorMeta.value_type_name())


class ProcessorTest(TestCase):
    def test_class_props(self):
        self.assertEqual("processor", Processor.value_name())
        self.assertEqual(
            "Processor | Type[ProcessorOp] | dict | str", Processor.value_type_name()
        )

    def test_define_processor(self):
        registry = {}

        class MyProcessorOp(ProcessorOp):
            def preprocess(
                self, file_path: str, opener_options: dict[str, Any]
            ) -> list[tuple[xr.Dataset, str]]:
                return []

            def postprocess(
                self, messages: list[list[Message]], file_path: str
            ) -> list[Message]:
                return []

        processor = define_processor(op_class=MyProcessorOp, registry=registry)

        self.assertTrue(hasattr(MyProcessorOp, "meta"))
        # noinspection PyUnresolvedReferences
        meta = MyProcessorOp.meta
        self.assertIsInstance(meta, ProcessorMeta)
        self.assertEqual("my-processor-op", meta.name)
        processor2: Processor = registry.get("my-processor-op")
        self.assertIs(processor, processor2)

    def test_define_processor_as_decorator(self):
        registry = {}

        @define_processor(registry=registry)
        class MyProcessorOp(ProcessorOp):
            def preprocess(
                self, file_path: str, opener_options: dict[str, Any]
            ) -> list[tuple[xr.Dataset, str]]:
                return []

            def postprocess(
                self, messages: list[list[Message]], file_path: str
            ) -> list[Message]:
                return []

        self.assertTrue(hasattr(MyProcessorOp, "meta"))
        # noinspection PyUnresolvedReferences
        meta = MyProcessorOp.meta
        self.assertIsInstance(meta, ProcessorMeta)
        self.assertEqual("my-processor-op", meta.name)
        processor: Processor = registry.get("my-processor-op")
        self.assertIsInstance(processor, Processor)
        self.assertIs(MyProcessorOp, processor.op_class)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def test_define_processor_as_decorator_fail(self):
        with pytest.raises(
            TypeError,
            match=(
                "decorated processor component must be a subclass of ProcessorOp,"
                " but got MyProcessorOp"
            ),
        ):

            @define_processor()
            class MyProcessorOp:
                pass

    def test_define_processor_with_plugin(self):
        plugin = new_plugin(name="my-plugin")

        @plugin.define_processor()
        class MyProcessorOp(ProcessorOp):
            def preprocess(
                self, file_path: str, opener_options: dict[str, Any]
            ) -> list[tuple[xr.Dataset, str]]:
                return []

            def postprocess(
                self, messages: list[list[Message]], file_path: str
            ) -> list[Message]:
                return []

        self.assertTrue(hasattr(MyProcessorOp, "meta"))
        # noinspection PyUnresolvedReferences
        meta = MyProcessorOp.meta
        self.assertIsInstance(meta, ProcessorMeta)
        self.assertEqual("my-processor-op", meta.name)
        processor: Processor = plugin.processors.get("my-processor-op")
        self.assertIsInstance(processor, Processor)
        self.assertIs(MyProcessorOp, processor.op_class)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import DatasetNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "grid-mappings",
    version="1.0.0",
    type="problem",
    description=(
        "Grid mappings, if any, shall have valid grid mapping coordinate variables."
    ),
    docs_url=(
        "https://cfconventions.org/cf-conventions/cf-conventions.html"
        "#grid-mappings-and-projections"
    ),
)
class GridMappings(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        dataset = node.dataset

        # Get the mapping of grid mapping names to grid-mapped variables
        grid_mapped_vars = {
            str(v.attrs.get("grid_mapping")): str(k)
            for k, v in dataset.data_vars.items()
            if "grid_mapping" in v.attrs
        }

        if len(grid_mapped_vars) == 0:
            return

        # Check validity of grid mappings
        for gm_name, var_name in grid_mapped_vars.items():
            gm_var = dataset.variables.get(gm_name)
            if gm_var is None:
                ctx.report(
                    f"Missing grid mapping variable {gm_name!r}"
                    f" referred to by variable {var_name!r}."
                )
            else:
                if gm_name not in dataset.coords:
                    ctx.report(
                        f"Grid mapping variable {gm_name!r} should"
                        f" be a coordinate variable, not data variable."
                    )
                if gm_var.dims:
                    dims_text = ",".join(str(d) for d in gm_var.dims)
                    ctx.report(
                        f"Grid mapping variable {gm_name!r} should be a scalar,"
                        f" but has dimension(s) {dims_text}."
                    )
                # Note: we could check if creating a CRS from
                #   gm_var.attrs is possible using pyproj.CRS.from_cf().
                #   Report otherwise.
                grid_mapping_name = gm_var.attrs.get("grid_mapping_name")
                if not grid_mapping_name:
                    ctx.report(
                        f"Grid mapping variable {gm_name!r} is missing"
                        f" a valid attribute 'grid_mapping_name'."
                    )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Any
from unittest import TestCase

import xarray as xr

from xrlint.config import Config, ConfigObject
from xrlint.constants import CORE_PLUGIN_NAME, DATASET_ROOT_NAME
from xrlint.linter import Linter, new_linter
from xrlint.node import AttrNode, AttrsNode, DatasetNode, DataTreeNode, VariableNode
from xrlint.plugin import new_plugin
from xrlint.processor import ProcessorOp
from xrlint.result import Message, Result
from xrlint.rule import RuleContext, RuleExit, RuleOp


class LinterTest(TestCase):
    def test_default_config_is_empty(self):
        linter = Linter()
        self.assertEqual(Config(), linter.config)

    def test_new_linter(self):
        linter = new_linter()
        self.assertIsInstance(linter, Linter)
        self.assertEqual(1, len(linter.config.objects))
        config_obj = linter.config.objects[0]
        self.assertIsInstance(config_obj.plugins, dict)
        self.assertEqual({CORE_PLUGIN_NAME}, set(config_obj.plugins.keys()))
        self.assertEqual(None, config_obj.rules)

    def test_new_linter_recommended(self):
        linter = new_linter("recommended")
        self.assertIsInstance(linter, Linter)
        self.assertEqual(2, len(linter.config.objects))
        config_obj_0 = linter.config.objects[0]
        config_obj_1 = linter.config.objects[1]
        self.assertIsInstance(config_obj_0.plugins, dict)
        self.assertEqual({CORE_PLUGIN_NAME}, set(config_obj_0.plugins.keys()))
        self.assertIsInstance(config_obj_1.rules, dict)
        self.assertIn("coords-for-dims", config_obj_1.rules)

    def test_new_linter_all(self):
        linter = new_linter("all")
        self.assertIsInstance(linter, Linter)
        self.assertEqual(2, len(linter.config.objects))
        config_obj_0 = linter.config.objects[0]
        config_obj_1 = linter.config.objects[1]
        self.assertIsInstance(config_obj_0.plugins, dict)
        self.assertEqual({CORE_PLUGIN_NAME}, set(config_obj_0.plugins.keys()))
        self.assertIsInstance(config_obj_1.rules, dict)
        self.assertIn("coords-for-dims", config_obj_1.rules)


class LinterValidateWithConfigTest(TestCase):
    def test_config_with_config_list(self):
        linter = new_linter()
        result = linter.validate(
            xr.Dataset(),
            config=Config.from_value([{"rules": {"no-empty-attrs": 2}}]),
        )
        self.assert_result_ok(result, "Missing metadata, attributes are empty.")

    def test_config_with_list_of_config(self):
        linter = new_linter()
        result = linter.validate(
            xr.Dataset(),
            config=[{"rules": {"no-empty-attrs": 2}}],
        )
        self.assert_result_ok(result, "Missing metadata, attributes are empty.")

    def test_config_with_config_obj(self):
        linter = new_linter()
        result = linter.validate(
            xr.Dataset(),
            config={"rules": {"no-empty-attrs": 2}},
        )
        self.assert_result_ok(result, "Missing metadata, attributes are empty.")

    def test_config_with_file_not_found(self):
        linter = new_linter({"rules": {"no-empty-attrs": 2}})
        result = linter.validate("cube.nc")
        self.assert_result_ok(result, "No such file or directory:")

    def test_no_config(self):
        linter = Linter()
        result = linter.validate(xr.Dataset())
        self.assert_result_ok(result, "No configuration given or matches '<dataset>'.")

    def assert_result_ok(self, result: Result, expected_message: str):
        self.assertIsInstance(result, Result)
        self.assertEqual(1, len(result.messages))
        self.assertEqual(2, result.messages[0].severity)
        self.assertIn(expected_message, result.messages[0].message)


class LinterValidateTest(TestCase):
    # noinspection PyUnusedLocal
    def setUp(self):
        plugin = new_plugin(name="test")

        @plugin.define_rule("no-space-in-attr-name")
        class AttrVer(RuleOp):
            def validate_attr(self, ctx: RuleContext, node: AttrNode):
                if " " in node.name:
                    ctx.report(f"Attribute name with space: {node.name!r}")

        @plugin.define_rule("no-empty-attrs")
        class AttrsVer(RuleOp):
            def validate_attrs(self, ctx: RuleContext, node: AttrsNode):
                if not node.attrs:
                    ctx.report("Empty attributes")

        @plugin.define_rule("data-var-dim-must-have-coord")
        class VariableVer(RuleOp):
            def validate_variable(self, ctx: RuleContext, node: VariableNode):
                if node.in_data_vars():
                    for dim_name in node.array.dims:
                        if dim_name not in ctx.dataset.coords:
                            ctx.report(
                                f"Dimension {dim_name!r}"
                                f" of data variable {node.name!r}"
                                f" is missing a coordinate variable"
                            )

        @plugin.define_rule("dataset-without-data-vars")
        class DatasetVer(RuleOp):
            def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
                if len(node.dataset.data_vars) == 0:
                    ctx.report("Dataset does not have data variables")
                    raise RuleExit  # no need to traverse further

        @plugin.define_rule("datatree-without-data-vars")
        class DataTreeVer(RuleOp):
            def validate_datatree(self, ctx: RuleContext, node: DataTreeNode):
                if len(node.datatree.data_vars) == 0:
                    ctx.report("DataTree does not have data variables")

        @plugin.define_processor("multi-level-dataset")
        class MultiLevelDataset(ProcessorOp):
            def preprocess(
                self, file_path: str, _opener_options: dict[str, Any]
            ) -> list[tuple[xr.Dataset, str]]:
                if file_path == "bad.levels":
                    raise OSError("bad checksum")
                return [
                    (xr.Dataset(attrs={"title": "Level 0"}), file_path + "/0.zarr"),
                    (xr.Dataset(attrs={"title": "Level 1"}), file_path + "/1.zarr"),
                ]

            def postprocess(
                self, messages: list[list[Message]], file_path: str
            ) -> list[Message]:
                return messages[0] + messages[1]

        config = ConfigObject(plugins={"test": plugin})
        self.linter = Linter(config)
        super().setUp()

    def test_rules_are_ok(self):
        self.assertEqual(
            [
                "no-space-in-attr-name",
                "no-empty-attrs",
                "data-var-dim-must-have-coord",
                "dataset-without-data-vars",
                "datatree-without-data-vars",
            ],
            list(self.linter.config.objects[0].plugins["test"].rules.keys()),
        )

    def test_linter_respects_rule_severity_error(self):
        result = self.linter.validate(
            xr.Dataset(), rules={"test/dataset-without-data-vars": 2}
        )
        self.assertEqual(
            Result(
                config_object=result.config_object,
                file_path="<dataset>",
                messages=[
                    Message(
                        message="Dataset does not have data variables",
                        node_path=DATASET_ROOT_NAME,
                        rule_id="test/dataset-without-data-vars",
                        severity=2,
                    )
                ],
            ),
            result,
        )
        self.assertEqual(0, result.warning_count)
        self.assertEqual(1, result.error_count)
        self.assertEqual(0, result.fatal_error_count)

    def test_linter_respects_rule_severity_warn(self):
        result = self.linter.validate(
            xr.Dataset(), rules={"test/dataset-without-data-vars": 1}
        )
        self.assertEqual(
            Result(
                config_object=result.config_object,
                file_path="<dataset>",
                messages=[
                    Message(
                        message="Dataset does not have data variables",
                        node_path=DATASET_ROOT_NAME,
                        rule_id="test/dataset-without-data-vars",
                        severity=1,
                    )
                ],
            ),
            result,
        )
        self.assertEqual(1, result.warning_count)
        self.assertEqual(0, result.error_count)
        self.assertEqual(0, result.fatal_error_count)

    def test_linter_respects_rule_severity_off(self):
        result = self.linter.validate(
            xr.Dataset(), rules={"test/dataset-without-data-vars": 0}
        )
        self.assertEqual(
            Result(
                config_object=result.config_object,
                file_path="<dataset>",
                messages=[],
            ),
            result,
        )
        self.assertEqual(0, result.warning_count)
        self.assertEqual(0, result.error_count)
        self.assertEqual(0, result.fatal_error_count)

    def test_linter_recognized_unknown_rule(self):
        result = self.linter.validate(xr.Dataset(), rules={"test/dataset-is-fast": 2})
        self.assertEqual(
            [
                Message(
                    message="unknown rule 'test/dataset-is-fast'",
                    rule_id="test/dataset-is-fast",
                    node_path=DATASET_ROOT_NAME,
                    severity=2,
                    fatal=True,
                )
            ],
            result.messages,
        )

    def test_linter_recognized_datatree_rule(self):
        result = self.linter.validate(
            xr.DataTree(
                children={
                    "measurement": xr.DataTree(
                        children={
                            "r10m": xr.DataTree(),
                            "r20m": xr.DataTree(),
                            "r60m": xr.DataTree(),
                        }
                    )
                }
            ),
            rules={"test/datatree-without-data-vars": 2},
        )
        self.assertEqual(
            [
                Message(
                    message="DataTree does not have data variables",
                    node_path="dt",
                    rule_id="test/datatree-without-data-vars",
                    severity=2,
                    fatal=None,
                    fix=None,
                    suggestions=None,
                ),
                Message(
                    message="DataTree does not have data variables",
                    node_path="dt/measurement",
                    rule_id="test/datatree-without-data-vars",
                    severity=2,
                    fatal=None,
                    fix=None,
                    suggestions=None,
                ),
                Message(
                    message="DataTree does not have data variables",
                    node_path="dt/measurement/r10m",
                    rule_id="test/datatree-without-data-vars",
                    severity=2,
                ),
                Message(
                    message="DataTree does not have data variables",
                    node_path="dt/measurement/r20m",
                    rule_id="test/datatree-without-data-vars",
                    severity=2,
                ),
                Message(
                    message="DataTree does not have data variables",
                    node_path="dt/measurement/r60m",
                    rule_id="test/datatree-without-data-vars",
                    severity=2,
                ),
            ],
            result.messages,
        )
        self.assertEqual(0, result.warning_count)
        self.assertEqual(5, result.error_count)
        self.assertEqual(0, result.fatal_error_count)

    def test_linter_real_life_scenario(self):
        dataset = xr.Dataset(
            attrs={
                # issue #1: space in attr name
                "created at": "10:20"
            },
            data_vars={
                "chl": (
                    xr.DataArray(
                        [[[1, 2], [3, 4]]],
                        dims=["time", "y", "x"],
                        attrs={"units": "mg/m^-3"},
                    )
                ),
                # issue #2: attrs missing
                "tsm": xr.DataArray([[[1, 2], [3, 4]]], dims=["time", "y", "x"]),
            },
            coords={
                "x": xr.DataArray([0.1, 0.2], dims="x", attrs={"units": "m"}),
                "y": xr.DataArray([0.2, 0.3], dims="y", attrs={"units": "m"}),
                # issue #3 + #4: missing "time" coord
            },
        )
        dataset.encoding["source"] = "chl-tsm.zarr"

        result = self.linter.validate(
            dataset,
            config={
                "rules": {
                    "test/no-space-in-attr-name": "error",
                    "test/no-empty-attrs": "warn",
                    "test/data-var-dim-must-have-coord": "error",
                    "test/dataset-without-data-vars": "warn",
                },
            },
        )
        self.assertEqual(
            Result(
                config_object=result.config_object,
                file_path="chl-tsm.zarr",
                messages=[
                    Message(
                        message="Attribute name with space: 'created at'",
                        node_path=f"{DATASET_ROOT_NAME}.attrs['created at']",
                        rule_id="test/no-space-in-attr-name",
                        severity=2,
                    ),
                    Message(
                        message="Empty attributes",
                        node_path=f"{DATASET_ROOT_NAME}.data_vars['tsm'].attrs",
                        rule_id="test/no-empty-attrs",
                        severity=1,
                    ),
                    Message(
                        message=(
                            "Dimension 'time' of data "
                            "variable 'chl' is missing a "
                            "coordinate variable"
                        ),
                        node_path=f"{DATASET_ROOT_NAME}.data_vars['chl']",
                        rule_id="test/data-var-dim-must-have-coord",
                        severity=2,
                    ),
                    Message(
                        message=(
                            "Dimension 'time' of data "
                            "variable 'tsm' is missing a "
                            "coordinate variable"
                        ),
                        node_path=f"{DATASET_ROOT_NAME}.data_vars['tsm']",
                        rule_id="test/data-var-dim-must-have-coord",
                        severity=2,
                    ),
                ],
            ),
            result,
        )
        self.assertEqual(1, result.warning_count)
        self.assertEqual(3, result.error_count)
        self.assertEqual(0, result.fatal_error_count)

    def test_processor_ok(self):
        result = self.linter.validate(
            "test.levels",
            config={
                "processor": "test/multi-level-dataset",
                "rules": {"test/dataset-without-data-vars": "warn"},
            },
        )

        self.assertEqual(
            [
                Message(
                    message="Dataset does not have data variables",
                    node_path=f"{DATASET_ROOT_NAME}[0]",
                    rule_id="test/dataset-without-data-vars",
                    severity=1,
                ),
                Message(
                    message="Dataset does not have data variables",
                    node_path=f"{DATASET_ROOT_NAME}[1]",
                    rule_id="test/dataset-without-data-vars",
                    severity=1,
                ),
            ],
            result.messages,
        )

    def test_processor_fail(self):
        result = self.linter.validate(
            "bad.levels",
            config={
                "processor": "test/multi-level-dataset",
                "rules": {"test/dataset-without-data-vars": "warn"},
            },
        )

        self.assertEqual(
            [
                Message(
                    message="bad checksum",
                    severity=2,
                    fatal=True,
                    node_path=DATASET_ROOT_NAME,
                )
            ],
            result.messages,
        )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import DatasetNode
from xrlint.plugins.xcube.constants import TIME_NAME
from xrlint.plugins.xcube.plugin import plugin
from xrlint.plugins.xcube.util import get_dataset_level_info, is_spatial_var
from xrlint.rule import RuleContext, RuleOp
from xrlint.util.formatting import format_seq


@plugin.define_rule(
    "ml-dataset-time",
    version="1.0.0",
    type="problem",
    description=(
        "The `time` dimension of multi-level datasets should use a chunk size of 1."
        " This allows for faster image tile generation for visualisation."
    ),
    docs_url="https://xcube.readthedocs.io/en/latest/mldatasets.html#definition",
)
class MLDatasetTime(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        level_info = get_dataset_level_info(node.dataset)
        if level_info is None:
            # ok, this rules applies only to level datasets opened
            # by the xcube multi-level processor
            return

        if TIME_NAME not in node.dataset.sizes or node.dataset.sizes[TIME_NAME] <= 1:
            # ok, no time dimension used or no time extent
            return

        for var_name, var in node.dataset.data_vars.items():
            if is_spatial_var(var) and TIME_NAME in var.dims and var.chunks is not None:
                time_index = var.dims.index(TIME_NAME)
                time_chunks = var.chunks[time_index]
                if not all(c == 1 for c in time_chunks):
                    ctx.report(
                        f"Variable {var_name!r} uses chunking for {TIME_NAME!r}"
                        f" that differs from from one: {format_seq(time_chunks)}."
                    )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.util.merge import merge_arrays, merge_dicts, merge_set_lists, merge_values


class NamingTest(TestCase):
    def test_merge_arrays(self):
        self.assertEqual(None, merge_arrays(None, None))
        self.assertEqual(["x"], merge_arrays(["x"], None))
        self.assertEqual(["y"], merge_arrays(None, ["y"]))
        self.assertEqual(["y"], merge_arrays(["x"], ["y"]))
        self.assertEqual(["z", "y"], merge_arrays(["x", "y"], ["z"]))
        self.assertEqual(["y", "z"], merge_arrays(["x"], ["y", "z"]))

    def test_merge_dicts(self):
        self.assertEqual(None, merge_dicts(None, None))
        self.assertEqual({"x": 1}, merge_dicts({"x": 1}, None))
        self.assertEqual({"y": 2}, merge_dicts(None, {"y": 2}))
        self.assertEqual({"x": 1, "y": 2}, merge_dicts({"x": 1}, {"y": 2}))
        self.assertEqual({"x": 2}, merge_dicts({"x": 1}, {"x": 2}))
        self.assertEqual(
            {"x": 1, "y": 2, "z": 3}, merge_dicts({"x": 1, "y": 2}, {"z": 3})
        )

    def test_merge_set_lists(self):
        self.assertEqual(None, merge_set_lists(None, None))
        self.assertEqual(["y"], merge_set_lists(None, ["y"]))
        self.assertEqual(["x"], merge_set_lists(["x"], None))
        self.assertEqual(["x", "y"], merge_set_lists(["x"], ["y"]))
        self.assertEqual(["y", "x"], merge_set_lists(["y", "x"], ["x", "y"]))

    def test_merge_values(self):
        self.assertEqual(None, merge_values(None, None))
        self.assertEqual(["z", "y"], merge_values(["x", "y"], ["z"]))
        self.assertEqual({"x": 1, "y": 2}, merge_values({"x": 1}, {"y": 2}))
        self.assertEqual(1, merge_values("x", 1))
        self.assertEqual(2, merge_values(["x"], 2))
        self.assertEqual(3, merge_values({"x": 1}, 3))



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from abc import ABC
from dataclasses import dataclass
from typing import Any, Hashable, Union

import xarray as xr


@dataclass(frozen=True, kw_only=True)
class Node(ABC):
    """Abstract base class for nodes passed to the methods of a
    rule operation [xrlint.rule.RuleOp][]."""

    path: str
    """Node path. So users find where in the tree the issue occurred."""

    parent: Union["Node", None]
    """Node parent. `None` for root nodes."""


@dataclass(frozen=True, kw_only=True)
class XarrayNode(Node):
    """Base class for `xr.Dataset` nodes."""

    def in_coords(self) -> bool:
        """Return `True` if this node is in `xr.Dataset.coords`."""
        return ".coords[" in self.path

    def in_data_vars(self) -> bool:
        """Return `True` if this node is a `xr.Dataset.data_vars`."""
        return ".data_vars[" in self.path

    def in_root(self) -> bool:
        """Return `True` if this node is a direct child of the dataset."""
        return not self.in_coords() and not self.in_data_vars()


@dataclass(frozen=True, kw_only=True)
class DataTreeNode(XarrayNode):
    """DataTree node."""

    name: Hashable
    """The name of the datatree."""

    datatree: xr.DataTree
    """The `xarray.DataTree` instance."""


@dataclass(frozen=True, kw_only=True)
class DatasetNode(XarrayNode):
    """Dataset node."""

    name: Hashable
    """The name of the dataset."""

    dataset: xr.Dataset
    """The `xarray.Dataset` instance."""


@dataclass(frozen=True, kw_only=True)
class VariableNode(XarrayNode):
    """Variable node.
    Could be a coordinate or data variable.
    If you need to distinguish, you can use expression
    `node.name in ctx.dataset.coords`.
    """

    name: Hashable
    """The name of the variable."""

    array: xr.DataArray
    """The `xarray.DataArray` instance."""


@dataclass(frozen=True, kw_only=True)
class AttrsNode(XarrayNode):
    """Attributes node."""

    attrs: dict[str, Any]
    """Attributes dictionary."""


@dataclass(frozen=True, kw_only=True)
class AttrNode(XarrayNode):
    """Attribute node."""

    name: str
    """Attribute name."""

    value: Any
    """Attribute value."""



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import unittest
from unittest import TestCase

from xrlint.config import Config
from xrlint.rule import RuleOp
from xrlint.util.importutil import import_value


class ExamplesTest(TestCase):
    def test_plugin_config(self):
        config, _ = import_value(
            "examples.plugin_config", "export_config", factory=Config.from_value
        )
        self.assertIsInstance(config, Config)
        self.assertEqual(3, len(config.objects))

    def test_virtual_plugin_config(self):
        config, _ = import_value(
            "examples.virtual_plugin_config",
            "export_config",
            factory=Config.from_value,
        )
        self.assertIsInstance(config, Config)
        self.assertEqual(3, len(config.objects))

    def test_rule_testing(self):
        from examples.rule_testing import GoodTitle, GoodTitleTest

        self.assertTrue(issubclass(GoodTitle, RuleOp))
        self.assertTrue(issubclass(GoodTitleTest, unittest.TestCase))



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import contextlib
from typing import Any, Literal

import xarray as xr

from xrlint.config import ConfigObject
from xrlint.constants import DATASET_ROOT_NAME, SEVERITY_ERROR
from xrlint.node import Node
from xrlint.result import Message, Suggestion
from xrlint.rule import RuleContext


class RuleContextImpl(RuleContext):
    def __init__(
        self,
        config: ConfigObject,
        dataset: xr.Dataset | xr.DataTree,
        file_path: str,
        file_index: int | None,
        access_latency: float | None,
    ):
        assert isinstance(config, ConfigObject)
        assert isinstance(dataset, (xr.Dataset | xr.DataTree))
        assert isinstance(file_path, str)
        assert file_index is None or isinstance(file_index, int)
        assert access_latency is None or isinstance(access_latency, float)
        if isinstance(dataset, xr.DataTree):
            datatree = dataset
            dataset = None
            if datatree.is_leaf:
                dataset = datatree.dataset
                datatree = None
        else:
            datatree = None
        self._config = config
        self._datatree = datatree
        self._dataset = dataset
        self._file_path = file_path
        self._file_index = file_index
        self._access_latency = access_latency
        self.messages: list[Message] = []
        self.rule_id: str | None = None
        self.severity: Literal[1, 2] = SEVERITY_ERROR
        self.node: Node | None = None

    @property
    def config(self) -> ConfigObject:
        return self._config

    @property
    def settings(self) -> dict[str, Any]:
        return self._config.settings or {}

    @property
    def datatree(self) -> xr.DataTree | None:
        return self._datatree

    @property
    def dataset(self) -> xr.Dataset | None:
        return self._dataset

    @dataset.setter
    def dataset(self, value: xr.Dataset) -> None:
        self._dataset = value

    @property
    def file_path(self) -> str:
        return self._file_path

    @property
    def file_index(self) -> int | None:
        return self._file_index

    @property
    def access_latency(self) -> float | None:
        return self._access_latency

    def report(
        self,
        message: str,
        *,
        fatal: bool | None = None,
        suggestions: list[Suggestion | str] | None = None,
    ):
        suggestions = (
            [Suggestion.from_value(s) for s in suggestions] if suggestions else None
        )
        m = Message(
            message=message,
            fatal=fatal,
            suggestions=suggestions,
            rule_id=self.rule_id,
            node_path=self.node.path if self.node is not None else DATASET_ROOT_NAME,
            severity=self.severity,
        )
        self.messages.append(m)

    @contextlib.contextmanager
    def use_state(self, **new_state):
        old_state = {k: getattr(self, k) for k in new_state.keys()}
        try:
            for k, v in new_state.items():
                setattr(self, k, v)
            yield
        finally:
            for k, v in old_state.items():
                setattr(self, k, v)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.formatter import Formatter, FormatterOp, FormatterRegistry


class FormatterRegistryTest(TestCase):
    def test_decorator_sets_meta(self):
        registry = FormatterRegistry()

        @registry.define_formatter()
        class MyFormat(FormatterOp):
            def format(self, *args, **kwargs) -> str:
                """Dummy"""

        my_rule = registry.get("my-format")
        self.assertIsInstance(my_rule, Formatter)
        self.assertEqual("my-format", my_rule.meta.name)
        self.assertEqual(None, my_rule.meta.version)
        self.assertEqual(None, my_rule.meta.schema)

    def test_decorator_registrations(self):
        registry = FormatterRegistry()

        @registry.define_formatter("my-fmt-a")
        class MyFormat1(FormatterOp):
            def format(self, *args, **kwargs) -> str:
                """Dummy"""

        @registry.define_formatter("my-fmt-b")
        class MyFormat2(FormatterOp):
            def format(self, *args, **kwargs) -> str:
                """Dummy"""

        @registry.define_formatter("my-fmt-c")
        class MyFormat3(FormatterOp):
            def format(self, *args, **kwargs) -> str:
                """Dummy"""

        fmt_names = list(registry.keys())
        fmt1, fmt2, fmt3 = list(registry.values())
        self.assertEqual(["my-fmt-a", "my-fmt-b", "my-fmt-c"], fmt_names)
        self.assertIsInstance(fmt1, Formatter)
        self.assertIsInstance(fmt2, Formatter)
        self.assertIsInstance(fmt3, Formatter)
        self.assertIsNot(fmt2, fmt1)
        self.assertIsNot(fmt3, fmt1)
        self.assertIsNot(fmt3, fmt2)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

"""
This configuration example demonstrates how to
define and use "virtual" plugins. Such plugins
can be defined inside a configuration item.

You can use this example directly via the Python API by passing it's
exported configuration to an instance of the `Linter` class or use
the XRLint CLI:

```bash
xrlint -c examples/virtual_plugin_config.py <OPTIONS> <FILES>
```
"""

from xrlint.node import DatasetNode
from xrlint.rule import RuleContext, RuleOp, define_rule


@define_rule("good-title", description="Dataset title should be 'Hello World!'.")
class GoodTitle(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        good_title = "Hello World!"
        if node.dataset.attrs.get("title") != good_title:
            ctx.report(
                "Attribute 'title' wrong.",
                suggestions=[f"Rename it to {good_title!r}."],
            )


# Define more rules here...


def export_config():
    return [
        # Define and use "hello" plugin
        {
            "plugins": {
                "hello": {
                    "meta": {
                        "name": "hello",
                        "version": "1.0.0",
                    },
                    "rules": {
                        "good-title": GoodTitle,
                        # Add more rules here...
                    },
                    "configs": {
                        "recommended": [
                            {
                                "rules": {
                                    "hello/good-title": "warn",
                                    # Configure more rules here...
                                },
                            }
                        ],
                        # Add more configurations here...
                    },
                },
            }
        },
        # Use recommended settings from xrlint
        "recommended",
        # Use recommended settings from "hello" plugin
        "hello/recommended",
    ]



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import sys

import click

# Warning: do not import heavy stuff here, it can
# slow down commands like "xrlint --help" otherwise.
from xrlint.cli.constants import (
    DEFAULT_MAX_WARNINGS,
    DEFAULT_OUTPUT_FORMAT,
    DEFAULT_CONFIG_FILE_YAML,
)
from xrlint.version import version


@click.command(name="xrlint")
@click.option(
    "--no-config-lookup",
    "no_config_lookup",
    help="Disable use of default configuration files",
    is_flag=True,
)
@click.option(
    "--config",
    "-c",
    "config_path",
    help="Use this configuration instead of looking for a default configuration file",
    metavar="FILE",
)
@click.option(
    "--print-config",
    "inspect_path",
    help="Print the configuration for the given file",
    metavar="FILE",
)
@click.option(
    "--plugin",
    "plugin_specs",
    help=(
        "Specify plugins. MODULE is the name of Python module"
        " that defines an 'export_plugin()' function."
    ),
    metavar="MODULE",
    multiple=True,
)
@click.option(
    "--rule",
    "rule_specs",
    help=(
        "Specify rules. SPEC must have format"
        " '<rule-name>: <rule-config>' (note the space character)."
    ),
    metavar="SPEC",
    multiple=True,
)
@click.option(
    "-o",
    "--output-file",
    "output_file",
    help="Specify file to write report to",
    metavar="FILE",
)
@click.option(
    "-f",
    "--format",
    "output_format",
    help=f"Use a specific output format - default: {DEFAULT_OUTPUT_FORMAT}",
    default=DEFAULT_OUTPUT_FORMAT,
    metavar="NAME",
)
@click.option(
    "--color/--no-color",
    "color_enabled",
    default=True,
    help="Force enabling/disabling of color",
)
@click.option(
    "--max-warnings",
    "max_warnings",
    help=(
        f"Number of warnings to trigger nonzero exit code"
        f" - default: {DEFAULT_MAX_WARNINGS}"
    ),
    type=int,
    default=DEFAULT_MAX_WARNINGS,
    metavar="COUNT",
)
@click.option(
    "--init",
    "init_mode",
    help=f"Write initial configuration file '{DEFAULT_CONFIG_FILE_YAML}' and exit.",
    is_flag=True,
)
@click.argument("files", nargs=-1)
@click.version_option(version)
@click.help_option()
def main(
    no_config_lookup: bool,
    config_path: str | None,
    inspect_path: str | None,
    plugin_specs: tuple[str, ...],
    rule_specs: tuple[str, ...],
    max_warnings: int,
    output_file: str | None,
    output_format: str,
    color_enabled: bool,
    init_mode: bool,
    files: tuple[str, ...],
):
    """Validate the given dataset FILES.

    When executed, XRLint does the following three things:

    (1) Unless options '--no-config-lookup' or '--config' are used
    it searches for a default configuration file in the current working
    directory. Default configuration files are determined by their
    filename, namely 'xrlint_config.py' or 'xrlint-config.<format>',
    where <format> refers to the filename extensions
    'json', 'yaml', and 'yml'. A Python configuration file ('*.py'),
    is expected to provide XRLInt configuration from a function
    'export_config()', which may include custom plugins and rules.

    (2) It then validates each dataset in FILES against the configuration.
    The default dataset patters are '**/*.zarr' and '**/.nc'.
    FILES may comprise also directories or URLs. The supported URL
    protocols are the ones supported by xarray. Using remote
    protocols may require installing additional packages such as
    S3Fs (https://s3fs.readthedocs.io/) for the 's3' protocol.
    If a directory is provided that not matched by any file pattern,
    it will be traversed recursively.

    (3) The validation result is dumped to standard output if not otherwise
    stated by '--output-file'. The output format is 'simple' by default.
    Other inbuilt formats are 'json' and 'html' which you can specify
    using the '--format' option.

    Please refer to the documentation (https://bcdev.github.io/xrlint/)
    for more information.
    """
    from xrlint.cli.engine import XRLint

    if init_mode:
        XRLint.init_config_file()
        return

    cli_engine = XRLint(
        no_config_lookup=no_config_lookup,
        config_path=config_path,
        plugin_specs=plugin_specs,
        rule_specs=rule_specs,
        output_format=output_format,
        output_path=output_file,
        output_styled=color_enabled,
        max_warnings=max_warnings,
    )

    if inspect_path:
        cli_engine.init_config()
        cli_engine.print_config_for_file(inspect_path)
        return

    if files:
        cli_engine.init_config()
        results = cli_engine.validate_files(files)
        report = cli_engine.format_results(results)
        cli_engine.write_report(report)

        error_status = cli_engine.result_stats.error_count > 0
        max_warn_status = cli_engine.max_warnings_exceeded
        if max_warn_status and not error_status:
            click.echo("Maximum number of warnings exceeded.")
        if max_warn_status or error_status:
            raise click.exceptions.Exit(1)


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import platform
import re
from functools import cached_property
from typing import Literal

_WIN_OS = platform.system() == "Windows"


class FilePattern:
    """A file path pattern that naively implements a subset
    of the [minimatch](https://github.com/isaacs/minimatch)
    pattern specification.

    It currently supports:

    * `?` single character
    * `*` anything except slashes
    * `**` anything including slashes ("glob-star")
    * `!<pattern>` negations
    * `#<comment>` comments

    The actual pattern matching is implemented with
    Python regular expressions.

    Args:
        pattern: A "minimatch" pattern.
        flip_negate: Returns from negate expressions the same as
            if they were not negated, i.e.,
            `True` on a hit, `False` on a miss.
    """

    def __init__(self, pattern: str, flip_negate: bool = False):
        self._pattern = pattern
        self._flip_negate = flip_negate

        self._empty = False
        self._comment = False
        self._negate = False
        self._dir: Literal[True, None] = None  # we cannot know

        if not pattern:
            self._empty = True
        else:
            if pattern[0] == "#":
                self._comment = True
                pattern = pattern[1:]
            elif pattern[0] == "!":
                self._negate = True
                pattern = pattern[1:]
                while pattern and pattern[0] == "!":
                    self._negate = not self._negate
                    pattern = pattern[1:]
            while pattern and pattern[-1] == "/":
                self._dir = True
                pattern = pattern[:-1]
        self.__pattern = pattern

    @property
    def pattern(self) -> str:
        """Original pattern."""
        return self._pattern

    @property
    def empty(self) -> bool:
        """`True` if this matcher's pattern is empty."""
        return self._empty

    @property
    def comment(self) -> bool:
        """`True` if this matcher's pattern is a comment."""
        return self._comment

    @property
    def negate(self) -> bool:
        """`True` if this matcher's pattern negates."""
        return self._negate

    @property
    def dir(self) -> Literal[True, None]:
        """`True` if this matcher's pattern denotes a directory."""
        return self._dir

    @cached_property
    def _regex(self) -> re.Pattern:
        return _translate_to_regex(self.__pattern)

    def __str__(self):
        return self.pattern

    def __repr__(self):
        return f"FilePattern({self.pattern!r})"

    def __eq__(self, other):
        if other is self:
            return True
        if not isinstance(other, FilePattern):
            return False
        return self.pattern == other.pattern

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.pattern)

    def match(self, path: str) -> bool:
        """Match a file system path or URI against this pattern.

        Args:
            path: File system path or URI

        Returns:
            `True` if `path` matches.
        """
        if self._empty:
            return True
        if self._comment:
            return False

        path = path if not _WIN_OS else path.replace("\\", "")
        while path and path[-1] == "/":
            path = path[:-1]

        match_result = self._regex.match(path)
        if self._negate and not self._flip_negate:
            return match_result is None
        else:
            return match_result is not None


def _translate_to_regex(pattern: str) -> re.Pattern:
    """Translate the given
    [minimatch](https://github.com/isaacs/minimatch) pattern
    into a regex pattern.
    """
    # Escape all regex special characters except for * and ?
    pattern = re.escape(pattern)

    # Replace the escaped * and ? with their regex equivalents
    pattern = (
        pattern.replace(r"\*\*/", ".*/?")
        .replace(r"\*\*", ".*")
        .replace(r"\*", "[^/]*")
        .replace(r"\?", ".")
    )

    # Allow for trailing slashes, but don't force
    pattern = f"{pattern}/?"

    # Add start and end anchors to the pattern
    pattern = f"^{pattern}$"

    # Compile the pattern into a regex object
    return re.compile(pattern)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from collections.abc import Iterable
from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, Literal, Union

from xrlint.constants import (
    CORE_DOCS_URL,
    CORE_PLUGIN_NAME,
    SEVERITY_ERROR,
    SEVERITY_WARN,
)
from xrlint.util.constructible import ValueConstructible
from xrlint.util.serializable import JsonSerializable

if TYPE_CHECKING:  # pragma: no cover
    from xrlint.config import ConfigObject
    from xrlint.rule import RuleMeta


@dataclass(frozen=True, kw_only=True)
class EditInfo(JsonSerializable):
    """Not used yet."""


@dataclass(frozen=True)
class Suggestion(ValueConstructible, JsonSerializable):
    desc: str
    """Description of the suggestion."""

    data: dict[str, None] | None = None
    """Data that can be referenced in the description."""

    fix: EditInfo | None = None
    """Not used yet."""

    @classmethod
    def _from_str(cls, value: str, value_name: str | None = None) -> "Suggestion":
        return Suggestion(value)


@dataclass()
class Message(JsonSerializable):
    message: str
    """The error message."""

    node_path: str | None = None
    """Node path within the dataset.
    This property is None if the message does not
    apply to a certain dataset node.
    """

    rule_id: str | None = None
    """The rule name that generated this lint message.
    If this message is generated by the xrlint core
    rather than rules, this is None.
    """

    severity: Literal[1, 2] | None = None
    """The severity of this message.
    `1` means warning and `2` means error.
    """

    fatal: bool | None = None
    """True if this is a fatal error unrelated to a rule,
    like a parsing error.
    """

    fix: EditInfo | None = None
    """The EditInfo object of auto-fix.
    This property is None if this
    message is not fixable.

    Not used yet.
    """

    suggestions: list[Suggestion] | None = None
    """The list of suggestions. Each suggestion is the pair
    of a description and an EditInfo object to fix the dataset.
    API users such as editor integrations can choose one of them
    to fix the problem of this message.
    This property is None if this message does not have any suggestions.
    """


@dataclass(kw_only=True, frozen=True)
class Result(JsonSerializable):
    """The aggregated information of linting a dataset."""

    file_path: str
    """The absolute path to the file of this result.
    This is the string "<dataset>" if the file path is unknown
    (when you didn't pass the `file_path` option to the
    `xrlint.lint_dataset()` method).
    """

    config_object: Union["ConfigObject", None] = None
    """The configuration object that produced this result 
    together with `file_path`.
    """

    messages: list[Message] = field(default_factory=list)
    """The array of message objects."""

    @cached_property
    def warning_count(self) -> int:
        """The number of warnings. This includes fixable warnings."""
        return sum(1 if m.severity == SEVERITY_WARN else 0 for m in self.messages)

    @cached_property
    def error_count(self) -> int:
        """The number of errors. This includes fixable errors
        and fatal errors.
        """
        return sum(1 if m.severity == SEVERITY_ERROR else 0 for m in self.messages)

    @cached_property
    def fatal_error_count(self) -> int:
        """The number of fatal errors."""
        return sum(1 if m.fatal else 0 for m in self.messages)

    def to_html(self) -> str:
        from xrlint.formatters.html import format_result

        return "\n".join(format_result(self))

    def _repr_html_(self) -> str:
        return self.to_html()

    def get_docs_url_for_rule(self, rule_id: str) -> str | None:
        from xrlint.config import split_config_spec

        plugin_name, rule_name = split_config_spec(rule_id)
        if plugin_name == CORE_PLUGIN_NAME or plugin_name == "xcube":
            return f"{CORE_DOCS_URL}#{rule_name}"
        try:
            plugin = self.config_object.get_plugin(plugin_name)
            rule = self.config_object.get_rule(rule_id)
            return rule.meta.docs_url or plugin.meta.docs_url
        except ValueError:
            return None


def get_rules_meta_for_results(results: list[Result]) -> dict[str, "RuleMeta"]:
    """Get all rule metadata from the list of `results`.

    Args:
        results: List of `Result` objects.

    Returns:
        A dictionary that maps unique rule names to `RuleMeta` objects.
    """
    rules_meta = {}
    for result in results:
        for message in result.messages:
            if message.rule_id:
                rule = result.config_object.get_rule(message.rule_id)
                rules_meta[message.rule_id] = rule.meta
    return rules_meta


@dataclass()
class ResultStats:
    """Utility for collecting simple statistics from results."""

    error_count: int = 0
    warning_count: int = 0
    result_count: int = 0

    def collect(self, results: Iterable[Result]) -> Iterable[Result]:
        """Collect statistics from `results`."""
        for result in results:
            self.error_count += result.error_count
            self.warning_count += result.warning_count
            self.result_count += 1
            yield result



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).


def to_kebab_case(name: str) -> str:
    return _to_lc_case(name, "-")


def to_snake_case(name: str) -> str:
    return _to_lc_case(name, "_")


# noinspection SpellCheckingInspection
def _to_lc_case(name: str, sep: str) -> str:
    lc_case = []
    pl = False  # (p)revious character is (l)owercase
    pu = False  # (p)revious character is (u)ppercase
    ppu = False  # (p)pre-(p)revious character is (u)ppercase
    for cc in name:
        cl = cc.islower()  # (c)urrent character is (l)owercase
        cu = cc.isupper()  # (c)urrent character is (u)ppercase
        if not cc.isalnum():
            _append_sep(lc_case, sep)
        else:
            if pl and not cl:
                _append_sep(lc_case, sep)
            elif ppu and pu and cl:
                _insert_sep(lc_case, sep)
            lc_case.append(cc.lower() if not cl else cc)
        ppu = pu
        pu = cu
        pl = cl
    return "".join(lc_case)


def _append_sep(sc_name, sep):
    n = len(sc_name)
    if n > 0 and sc_name[-1] != sep:
        sc_name.append(sep)


def _insert_sep(sc_name, sep):
    n = len(sc_name)
    if n > 1 and sc_name[-2] != sep:
        sc_name.append(sc_name[-1])
        sc_name[-2] = sep



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import DatasetNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleExit, RuleOp


@plugin.define_rule(
    "no-empty-chunks",
    version="1.0.0",
    type="suggestion",
    description=(
        "Empty chunks should not be encoded and written."
        " The rule currently applies to Zarr format only."
    ),
    docs_url=(
        "https://docs.xarray.dev/en/stable/generated/xarray.Dataset.to_zarr.html"
        "#xarray-dataset-to-zarr"
    ),
)
class NoEmptyChunks(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        source = node.dataset.encoding.get("source")
        is_zarr = isinstance(source, str) and source.endswith(".zarr")
        if is_zarr:
            for var in node.dataset.data_vars.values():
                is_chunked_in_storage = (
                    "_FillValue" in var.encoding
                    and "chunks" in var.encoding
                    and tuple(var.encoding.get("chunks")) != tuple(var.shape)
                )
                if is_chunked_in_storage:
                    ctx.report("Consider writing with `write_empty_chunks=False`.")
                    break
        # no need to traverse further
        raise RuleExit



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np

from xrlint.node import VariableNode
from xrlint.plugins.xcube.plugin import plugin
from xrlint.rule import RuleContext, RuleExit, RuleOp
from xrlint.util.formatting import format_count, format_seq


@plugin.define_rule(
    "increasing-time",
    version="1.0.0",
    type="problem",
    description="Time coordinate labels should be monotonically increasing.",
    docs_url=(
        "https://xcube.readthedocs.io/en/latest/cubespec.html#temporal-reference"
    ),
)
class IncreasingTime(RuleOp):
    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        array = node.array
        if node.in_coords() and node.name == "time" and array.dims == ("time",):
            diff_array: np.ndarray = array.diff("time").values
            if not np.count_nonzero(diff_array > 0) == diff_array.size:
                check_indexes(ctx, diff_array == 0, "Duplicate")
                check_indexes(ctx, diff_array < 0, "Backsliding")
                raise RuleExit  # No need to apply rule any further


def check_indexes(ctx, cond: np.ndarray, issue_name: str):
    (indexes,) = np.nonzero(cond)
    if indexes.size > 0:
        index_text = format_count(indexes.size, singular="index", plural="indexes")
        ctx.report(
            f"{issue_name} 'time' coordinate label at {index_text}"
            f" {format_seq(indexes)}."
        )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np
import xarray as xr

nx = 720
ny = 360
nt = 5

ncx = 180
ncy = 180
nct = 1


def make_dataset() -> xr.Dataset:
    """Create a dataset that passes xrlint core rules."""

    return xr.Dataset(
        attrs=dict(
            Conventions="CF-1.10",
            title="SST-Climatology Subset",
            history="2025-01-31 17:31:00 - created;",
            institution="BC",
            source="SST CCI L4",
            references="https://climate.esa.int/en/projects/sea-surface-temperature/",
            comment="Demo dataset",
        ),
        coords={
            "x": xr.DataArray(
                np.linspace(-180, 180, nx),
                dims="x",
                attrs={
                    "standard_name": "longitude",
                    "long_name": "longitude",
                    "units": "degrees_east",
                },
            ),
            "y": xr.DataArray(
                np.linspace(-90, 90, ny),
                dims="y",
                attrs={
                    "standard_name": "latitude",
                    "long_name": "latitude",
                    "units": "degrees_north",
                },
            ),
            "time": xr.DataArray(
                [365 * i for i in range(nt)],
                dims="time",
                attrs={
                    "standard_name": "time",
                    "long_name": "time",
                    "units": "days since 2020-01-01 +0:00",
                    "calendar": "gregorian",
                },
            ),
            "spatial_ref": xr.DataArray(
                0,
                attrs={
                    "grid_mapping_name": "latitude_longitude",
                    "semi_major_axis": 6371000.0,
                    "inverse_flattening": 0,
                },
            ),
        },
        data_vars={
            "sst": xr.DataArray(
                np.random.random((nt, ny, nx)),
                dims=["time", "y", "x"],
                attrs={
                    "standard_name": "sea_surface_temperature",
                    "long_name": "sea surface temperature",
                    "units": "kelvin",
                    "grid_mapping": "spatial_ref",
                },
            ),
            "sst_anomaly": xr.DataArray(
                np.random.random((nt, ny, nx)),
                dims=["time", "y", "x"],
                attrs={
                    "standard_name": "sea_surface_temperature_anomaly",
                    "long_name": "sea surface temperature anomaly",
                    "units": "kelvin",
                    "grid_mapping": "spatial_ref",
                },
            ),
        },
    ).chunk(time=nct, y=ncy, x=ncx)


def make_dataset_with_issues() -> xr.Dataset:
    """Create a dataset that produces issues with xrlint core rules."""
    invalid_ds = make_dataset()
    invalid_ds.x.attrs["units"] = "degrees"
    invalid_ds.x.attrs["axis"] = "x"
    del invalid_ds.y.attrs["standard_name"]
    invalid_ds.y.attrs["axis"] = "y"
    invalid_ds.time.attrs["units"] = "days since 2020-01-01 UTC"
    invalid_ds.attrs = {}
    invalid_ds.sst.attrs["units"] = 1
    invalid_ds["sst_avg"] = xr.DataArray(
        np.random.random((nx, ny)), dims=["x", "y"], attrs={"units": "kelvin"}
    )
    return invalid_ds



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import xarray as xr

from tests.plugins.xcube.helpers import make_cube_levels
from xrlint.plugins.xcube.rules.ml_dataset_xy import MLDatasetXY
from xrlint.plugins.xcube.util import LevelsMeta, get_dataset_level_info
from xrlint.testing import RuleTest, RuleTester

meta = LevelsMeta(
    version="1.0",
    num_levels=4,
    use_saved_levels=True,
    agg_methods={"chl": "mean"},
)
levels = make_cube_levels(4, 720, 360, nt=3, meta=meta)

valid_dataset_0 = levels[0]
valid_dataset_1 = levels[1]
valid_dataset_2 = levels[2]
valid_dataset_3 = levels[3]
valid_dataset_4 = xr.Dataset()

levels = make_cube_levels(4, 720, 360, meta=meta)
valid_dataset_5 = levels[2]
level_info = get_dataset_level_info(valid_dataset_5)
for ds, _ in level_info.datasets:
    # remove spatial vars
    del ds["chl"]

levels = make_cube_levels(4, 720, 360, meta=meta)
invalid_dataset_0 = levels[2].copy()
# simulate resolution mismatch by exchanging 2 levels
level_info = get_dataset_level_info(invalid_dataset_0)
level_datasets = level_info.datasets
level_0_dataset = level_datasets[0]
level_1_dataset = level_datasets[1]
level_info.datasets[0] = level_1_dataset
level_info.datasets[1] = level_0_dataset


MLDatasetXYTest = RuleTester.define_test(
    "ml-dataset-xy",
    MLDatasetXY,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
        RuleTest(dataset=valid_dataset_3),
        RuleTest(dataset=valid_dataset_4),
        RuleTest(dataset=valid_dataset_5),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=2),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from collections.abc import Hashable
from dataclasses import dataclass

import xarray as xr

from xrlint.util.constructible import MappingConstructible

from .constants import LAT_NAME, LON_NAME, ML_INFO_ATTR, X_NAME, Y_NAME


@dataclass(frozen=True, kw_only=True)
class LevelsMeta(MappingConstructible):
    """The contents of a xcube `.zlevels` meta-info file."""

    version: str
    num_levels: int
    tile_size: tuple[int, int] | None = None
    use_saved_levels: bool | None = None
    agg_methods: dict[str, str] | None = None


@dataclass(frozen=True, kw_only=True)
class LevelInfo:
    """Level info added to each level dataset to allow for validation
    of multi-level datasets.

    Note, this need to be aligned with
    <https://xcube.readthedocs.io/en/latest/mldatasets.html#the-xcube-levels-format>
    """

    level: int
    """Current level index."""

    num_levels: int
    """Actual number of levels found."""

    datasets: list[tuple[xr.Dataset, str]]
    """A list of num_levels pairs comprising
     level dataset and level file path.
     """

    meta: LevelsMeta | None = None
    """Content of a `.zlevels` file, if file was found."""


def get_dataset_level_info(dataset: xr.Dataset) -> LevelInfo | None:
    return dataset.attrs.get(ML_INFO_ATTR)


def set_dataset_level_info(dataset: xr.Dataset, level_info: LevelInfo):
    dataset.attrs[ML_INFO_ATTR] = level_info


def attach_dataset_level_infos(
    level_datasets: list[tuple[xr.Dataset, str]],
    meta: LevelsMeta | None = None,
):
    for level, (level_dataset, _) in enumerate(level_datasets):
        set_dataset_level_info(
            level_dataset,
            LevelInfo(
                level=level,
                num_levels=len(level_datasets),
                meta=meta,
                datasets=level_datasets,
            ),
        )


def is_spatial_var(var: xr.DataArray) -> bool:
    """Return 'True' if `var` looks like a spatial 2+d variable."""
    if var.ndim < 2:
        return False
    y_name, x_name = var.dims[-2:]
    return (x_name == X_NAME and y_name == Y_NAME) or (
        x_name == LON_NAME and y_name == LAT_NAME
    )


def get_spatial_size(
    dataset: xr.Dataset,
) -> tuple[tuple[Hashable, int], tuple[Hashable, int]] | None:
    """Return (x_size, y_size) for given dataset."""
    for k, v in dataset.data_vars.items():
        if is_spatial_var(v):
            y_name, x_name = v.dims[-2:]
            x_size = dataset.sizes[x_name]
            y_size = dataset.sizes[y_name]
            if x_size and y_size:
                return (x_name, x_size), (y_name, y_size)
    return None


def resolve_path(path: str, root_path: str | None = None) -> str:
    abs_level_path = path
    if root_path is not None and not is_absolute_path(path):
        abs_level_path = f"{root_path}/{path}"
    parts = abs_level_path.rstrip("/").replace("\\", "/").split("/")
    return "/".join(
        p
        for i, p in enumerate(parts)
        if p not in (".", "..") and (i == len(parts) - 1 or parts[i + 1] != "..")
    )


def is_absolute_path(path: str) -> bool:
    return (
        # Unix abs path
        path.startswith("/")
        # URL
        or "://" in path
        # Windows abs paths
        or path.startswith("\\\\")
        or path.find(":\\", 1) == 1
        or path.find(":/", 1) == 1
    )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.formatters.html import Html, HtmlText

from .helpers import get_context, get_test_results


class HtmlTest(TestCase):
    def test_html(self):
        results = get_test_results()
        formatter = Html()
        text = formatter.format(
            context=get_context(),
            results=results,
        )
        self.assertIsInstance(text, HtmlText)
        self.assertIs(text, text._repr_html_())
        self.assertIn("</p>", text)

    def test_html_with_meta(self):
        results = get_test_results()
        formatter = Html(with_meta=True)
        text = formatter.format(
            context=get_context(),
            results=results,
        )
        self.assertIsInstance(text, HtmlText)
        self.assertIs(text, text._repr_html_())
        self.assertIn("</p>", text)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import itertools
import json
import re
from typing import Any

import fsspec
import xarray as xr

from xrlint.plugins.xcube.constants import ML_FILE_PATTERN, ML_META_FILENAME
from xrlint.plugins.xcube.plugin import plugin
from xrlint.plugins.xcube.util import (
    LevelsMeta,
    attach_dataset_level_infos,
    resolve_path,
)
from xrlint.processor import ProcessorOp
from xrlint.result import Message

level_pattern = re.compile(r"^(\d+)(?:\.zarr)?$")
link_pattern = re.compile(r"^(\d+)(?:\.link)?$")


@plugin.define_processor("multi-level-dataset")
class MultiLevelDatasetProcessor(ProcessorOp):
    f"""This processor should be used with `files: [{ML_FILE_PATTERN}"]`."""

    def preprocess(
        self, file_path: str, opener_options: dict[str, Any]
    ) -> list[tuple[xr.Dataset | xr.DataTree, str]]:
        fs, fs_path = get_filesystem(file_path, opener_options)

        file_names = [
            # extracting the filename could be done more robustly
            f.replace(fs_path, "").strip("/")
            for f in fs.listdir(fs_path, detail=False)
        ]

        # check for optional ".zlevels" that provides meta-info
        meta = None
        if ML_META_FILENAME in file_names:
            with fs.open(f"{fs_path}/{ML_META_FILENAME}") as stream:
                meta = LevelsMeta.from_value(json.load(stream))

        # check for optional ".zgroup"
        # if ".zgroup" in file_names:
        #     with fs.open(f"{fs_path}/.zgroup") as stream:
        #         group_props = json.load(stream)

        level_paths, num_levels = parse_levels(fs, file_path, file_names)

        engine = opener_options.pop("engine", "zarr")

        level_datasets: list[xr.Dataset | None] = []
        for level, level_path in level_paths.items():
            level_dataset = xr.open_dataset(level_path, engine=engine, **opener_options)
            level_datasets.append((level_dataset, level_path))

        attach_dataset_level_infos(level_datasets, meta=meta)

        return level_datasets

    def postprocess(
        self, messages: list[list[Message]], file_path: str
    ) -> list[Message]:
        return list(itertools.chain(*messages))


def get_filesystem(file_path: str, opener_options: dict[str, Any]):
    storage_options = (
        opener_options.get(
            "storage_options",
            opener_options.get("backend_kwargs", {}).get("storage_options"),
        )
        or {}
    )
    _fs, fs_path = fsspec.core.url_to_fs(file_path, **storage_options)
    fs: fsspec.AbstractFileSystem = _fs
    fs_path: str = fs_path.replace("\\", "/")
    return fs, fs_path


def parse_levels(
    fs: fsspec.AbstractFileSystem, dataset_path: str, file_names: list[str]
) -> tuple[dict[int, str], int]:
    level_paths: dict[int, str] = {}
    for file_name in file_names:
        # check for optional "<level>.link" that locates a level somewhere else
        m = link_pattern.match(file_name)
        if m is not None:
            level = int(m.group(1))
            link_path = fs.read_text(f"{dataset_path}/{file_name}")
            level_paths[level] = resolve_path(link_path, root_path=dataset_path)
        # check for regular "<level>.zarr"
        m = level_pattern.match(file_name)
        if m is not None:
            level = int(m.group(1))
            level_paths[level] = f"{dataset_path}/{file_name}"

    if not level_paths:
        raise ValueError("empty multi-level dataset")

    num_levels = max(level_paths.keys()) + 1
    for level in range(num_levels):
        if level not in level_paths:
            raise ValueError(
                f"missing dataset for level {level} in multi-level dataset"
            )

    return level_paths, num_levels



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import os
from contextlib import contextmanager


@contextmanager
def text_file(file_path: str, content: str):
    with open(file_path, mode="w") as f:
        f.write(content)
    try:
        yield file_path
    finally:
        os.remove(file_path)
